"""Functionality around Notion pages."""

from __future__ import annotations

from collections.abc import Iterator, Mapping
from typing import TYPE_CHECKING, Any, cast

from typing_extensions import Self, TypeIs

from ultimate_notion.blocks import ChildrenMixin, CommentMixin, DataObject, wrap_icon
from ultimate_notion.comment import Discussion
from ultimate_notion.core import NotionEntity, WorkspaceType, get_active_session, get_repr
from ultimate_notion.emoji import CustomEmoji, Emoji
from ultimate_notion.file import AnyFile, ExternalFile, NotionFile
from ultimate_notion.obj_api import blocks as obj_blocks
from ultimate_notion.obj_api import objects as objs
from ultimate_notion.obj_api import props as obj_props
from ultimate_notion.obj_api.core import raise_unset
from ultimate_notion.obj_api.props import MAX_ITEMS_PER_PROPERTY
from ultimate_notion.props import PropertyValue, Title
from ultimate_notion.rich_text import Text, render_md
from ultimate_notion.schema import Property
from ultimate_notion.templates import page_html
from ultimate_notion.utils import SList, is_notebook

if TYPE_CHECKING:
    from ultimate_notion.database import Database
    from ultimate_notion.schema import Schema


class PagePropertiesNS(Mapping[str, Any]):
    """Namespace of the properties of a page as defined in the schema of the database.

    This defines the `.props` namespace of a page `page` and updates the content
    on the Notion server side in case of an assignment.
    Access the properties with `page.props.property_name` or `page.props['Property Name']`.
    You can also convert it to a dictionary with `dict(page.props)`.
    """

    def __init__(self, page: Page, schema: type[Schema] | None) -> None:
        self._page = page
        self._schema = schema

    @property
    def _obj_prop_vals(self) -> dict[str, obj_props.PropertyValue]:
        """Return the low-level page properties"""
        return self._page.obj_ref.properties

    def _get_property(self, prop_name: str) -> PropertyValue:
        prop = self._obj_prop_vals.get(prop_name)

        if prop is None:
            msg = f'No such property: {prop_name}'
            raise AttributeError(msg)

        def has_len_max_mention(prop: obj_props.RichText | obj_props.Title) -> bool:
            return len(Text.wrap_obj_ref(prop.value).mentions) == MAX_ITEMS_PER_PROPERTY

        match prop:
            case obj_props.Relation() if prop.has_more:
                prop.has_more = False
                return self._page._get_property(prop_name)
            case obj_props.People() if len(prop.people) == MAX_ITEMS_PER_PROPERTY and not prop._is_retrieved:
                return self._page._get_property(prop_name)
            case obj_props.RichText() | obj_props.Title() if has_len_max_mention(prop) and not prop._is_retrieved:
                return self._page._get_property(prop_name)
            case _:
                return PropertyValue.wrap_obj_ref(prop)

    def __getitem__(self, prop_name: str) -> Any:
        return self._get_property(prop_name).value

    def __setitem__(self, prop_name: str, value: Any) -> None:
        # Todo: use the schema of the database to see which properties are writeable at all.
        db = self._page.parent_db

        if db is None:
            msg = f'Trying to set a property but page {self._page} is not bound to any database'
            raise RuntimeError(msg)

        if not isinstance(value, PropertyValue):
            # construct concrete PropertyValue using the schema
            value = db.schema.get_prop(prop_name).prop_value(value)

        session = get_active_session()
        # update the property on the server (which will update the local data)
        session.api.pages.update(self._page.obj_ref, **{prop_name: value.obj_ref})

    def _check_schema(self) -> None:
        if self._schema is None:
            msg = f'{self.__class__.__name__} object is not bound to any schema'
            raise RuntimeError(msg)

    def __getattr__(self, attr_name: str) -> Any:
        self._check_schema()
        prop = cast(Property, getattr(self._schema, attr_name))
        return self[prop.name]

    def __setattr__(self, attr_name: str, value: Any) -> None:
        if attr_name.startswith('_'):
            return super().__setattr__(attr_name, value)

        self._check_schema()
        prop = cast(Property, getattr(self._schema, attr_name))
        self[prop.name] = value

    def __iter__(self) -> Iterator[str]:
        """Iterator of property names."""
        yield from self._obj_prop_vals.keys()

    def __len__(self) -> int:
        """Return the number of properties."""
        return len(self._obj_prop_vals)

    def __repr__(self) -> str:
        return repr(dict(self))

    def __str__(self) -> str:
        return str(dict(self))


class Page(
    ChildrenMixin[obj_blocks.Page], CommentMixin[obj_blocks.Page], DataObject[obj_blocks.Page], wraps=obj_blocks.Page
):
    """A Notion page.

    Attributes:
        props: accessor for all page properties
    """

    props: PagePropertiesNS

    @classmethod
    def wrap_obj_ref(cls, obj_ref: obj_blocks.Page, /) -> Self:
        obj = super().wrap_obj_ref(obj_ref)
        obj.props = obj._create_page_props_ns()
        return obj

    def _create_page_props_ns(self) -> PagePropertiesNS:
        """Create a namespace for the properties of this page defind by the database."""
        schema = None if self.parent_db is None else self.parent_db.schema
        return PagePropertiesNS(page=self, schema=schema)

    def __str__(self) -> str:
        return str(self.title)

    def __repr__(self) -> str:
        return get_repr(self, desc=self.title)

    def _repr_html_(self) -> str:  # noqa: PLW3201
        """Called by JupyterLab automatically to display this page."""
        return self.to_html()

    @property
    def is_page(self) -> bool:
        """Return whether the object is a page."""
        return True

    @property
    def is_locked(self) -> bool:
        """Return whether the page is locked for editing."""
        return raise_unset(self.obj_ref.is_locked)

    @property
    def url(self) -> str:
        """Return the URL of this page."""
        return raise_unset(self.obj_ref.url)

    @property
    def public_url(self) -> str | None:
        """Return the public URL of this database."""
        return self.obj_ref.public_url

    @property
    def subpages(self) -> list[Page]:
        """Return all contained pages within this page"""
        return [block for block in self.children if is_page_guard(block)]

    @property
    def subdbs(self) -> list[Database]:
        """Return all contained databases within this page"""
        return [block for block in self.children if is_db_guard(block)]

    @property
    def parent_db(self) -> Database | None:
        """If this page is located in a database return the database or None otherwise.

        This is a convenience method to avoid the need to check and cast the type of the parent.
        """

        if is_db_guard(self.parent):
            return self.parent
        else:
            return None

    @property
    def in_db(self) -> bool:
        """Return True if this page is located in a database."""
        return self.parent_db is not None

    @property
    def title(self) -> Text | None:
        """Title of the page."""
        if not (title := self.obj_ref.title):
            return None
        return Text.wrap_obj_ref(title)

    @title.setter
    def title(self, text: str | None) -> None:
        """Set the title of the page."""
        if text is None:
            text = ''
        title = Title(text)
        title_prop_name = self.obj_ref._get_title_prop_name()
        session = get_active_session()
        session.api.pages.update(self.obj_ref, **{title_prop_name: title.obj_ref})

    @property
    def icon(self) -> NotionFile | ExternalFile | Emoji | CustomEmoji | None:
        """Icon of the page, i.e. emojis, Notion's icons, or custom images."""
        if (icon := self.obj_ref.icon) is None:
            return None
        else:
            return wrap_icon(icon)

    @icon.setter
    def icon(self, icon: AnyFile | Emoji | CustomEmoji | str | None) -> None:
        """Set the icon of this page."""
        if isinstance(icon, str) and not isinstance(icon, Emoji | CustomEmoji):
            icon = Emoji(icon)
        icon_obj = None if icon is None else icon.obj_ref
        session = get_active_session()
        session.api.pages.set_attr(self.obj_ref, icon=icon_obj)

    @property
    def cover(self) -> ExternalFile | NotionFile | None:
        """Cover of the page."""
        match cover := self.obj_ref.cover:
            case objs.ExternalFile():
                return ExternalFile.wrap_obj_ref(cover)
            case objs.HostedFile():
                return NotionFile.wrap_obj_ref(cover)
            case None:
                return None
            case _:
                msg = f'Unknown cover object of type {type(cover)}.'
                raise RuntimeError(msg)

    @cover.setter
    def cover(self, cover: AnyFile | None) -> None:
        """Set the cover fo this page."""
        cover_obj = None if cover is None else cover.obj_ref
        session = get_active_session()
        session.api.pages.set_attr(self.obj_ref, cover=cover_obj)

    @property
    def comments(self) -> Discussion:
        """Return the discussion thread of this page.

        A page can only have a single discussion thread in contrast to inline comments.

        !!! note

            This functionality requires that your integration was granted *read* comment capabilities.
        """
        if not self._discussions:  # create an empty discussion thread
            self._comments = [Discussion([], parent=self)]
        return SList(self._discussions).item()

    def _get_property(self, prop_name: str) -> PropertyValue:
        session = get_active_session()
        prop_obj = self.props._obj_prop_vals[prop_name]
        prop_items = session.api.pages.properties.retrieve(self.obj_ref, prop_obj)
        prop_values = (item.value for item in prop_items)

        if isinstance(prop_obj, obj_props.PAGINATED_PROP_VALS):
            prop_obj.value = list(prop_values)
        else:  # we should have only one property-item in the iterator
            prop_obj.value = SList(prop_values).item()  # guaranteed to have only one item

        prop_obj._is_retrieved = True
        return PropertyValue.wrap_obj_ref(prop_obj)

    def get_property(self, prop_name: str) -> Any:
        """Directly retrieve the property value from the API.

        Use this method only if you want to retrieve a specific property value that
        might have been updated on the server side without reloading the whole page.
        In all other cases, use the `props` namespace of the page to avoid unnecessary API calls.
        """
        return self._get_property(prop_name).value

    def to_markdown(self) -> str:
        """Return the content of the page as Markdown.

        !!! note

            This will not include nested blocks, i.e. the children of top-level blocks.
        """
        md = f'# {self.title}\n\n'
        md += '\n'.join(block._to_markdown() for block in self.children)
        return md

    def _to_markdown(self) -> str:
        """Return the reference to this page as Markdown."""
        return f'[📄 **<u>{self.title}</u>**]({self.url})\n'

    def to_html(self, *, raw: bool = False) -> str:
        """Return the content of the page as HTML."""
        html = render_md(self.to_markdown())
        if not raw:
            html = page_html(html, title=self.title or 'Untitled Page')
        return html

    def show(self, *, simple: bool | None = None) -> None:
        """Show the content of the page, rendered in JupyterLab"""
        simple = simple if simple is not None else not is_notebook()
        md = self.to_markdown()

        if simple:
            print(md)  # noqa: T201
        else:
            from IPython.core.display import display_markdown  # noqa: PLC0415

            display_markdown(md, raw=True)

    def delete(self) -> Self:
        """Delete this page.

        !!! Warning

            Deleting a page will also delete all child pages and child databases recursively.
            If these objects are already cached in the session, they will not be updated.
            Use `session.cache.clear()` to clear the cache or call `reload()` on them.
        """
        if not self.is_deleted:
            session = get_active_session()
            session.api.pages.delete(self.obj_ref)
            self._delete_me_from_parent()
        return self

    def restore(self) -> Self:
        """Restore this page."""
        if self.is_deleted:
            session = get_active_session()
            session.api.pages.restore(self.obj_ref)
            if isinstance(self.parent, ChildrenMixin):
                self.parent._children = None  # forces a new retrieval of the parent's children next time
        return self

    def reload(self) -> Self:
        """Reload this page."""
        session = get_active_session()
        self.obj_ref = session.api.pages.retrieve(self.id)
        self._children = None  # forces a new retrieval of children next time
        self._comments = None  # forces a new retrieval of comments next time
        return self


def is_db_guard(obj: NotionEntity | WorkspaceType | None) -> TypeIs[Database]:
    """Return whether the object is a database as type guard."""
    return isinstance(obj, NotionEntity) and obj.is_db


def is_page_guard(obj: NotionEntity | WorkspaceType | None) -> TypeIs[Page]:
    """Return whether the object is a page as type guard."""
    return isinstance(obj, NotionEntity) and obj.is_page



"""Utilities to synchronize other services with Notion."""

from __future__ import annotations

import asyncio
import logging
import pickle  # noqa: S403
from abc import ABC, abstractmethod
from collections.abc import Generator
from datetime import datetime, timedelta
from enum import Enum
from typing import Any, TypeAlias

from pydantic import BaseModel, Field
from typing_extensions import Self

from ultimate_notion.config import get_cfg

_logger = logging.getLogger(__name__)

ID: TypeAlias = str


class ConflictMode(str, Enum):
    """Conflict resolution modes."""

    NEWER = 'newer'
    NOTION = 'notion'
    OTHER = 'other'  # any other service than Notion
    ERROR = 'error'


class State(BaseModel):
    """The state of a sync task.

    The state holds the synced objects and their attributes as a dictionary of Notion attributes."""

    ids: dict[ID, ID] = Field(default_factory=dict)
    """Maps Notion object ids to other object ids."""
    objs: dict[ID, dict[str, Any]] = Field(default_factory=dict)
    """Dictionary of Notion objects synced with other service and indexed by their ids."""


class SyncTask(ABC):
    """A task to be performed during a sync.

    This is an abstract base class to allow for different kinds of sync tasks.
    """

    _run_every_secs: float | None = None
    _in_total_times: int | None = None

    def __init__(
        self, name: str, attr_map: dict[str, str], conflict_mode: ConflictMode | str = ConflictMode.NEWER
    ) -> None:
        if isinstance(conflict_mode, str):
            conflict_mode = ConflictMode(conflict_mode)

        self.name = name
        self.attr_map = attr_map
        self.conflict_mode = conflict_mode
        self.state_path = get_cfg().ultimate_notion.sync_state_dir / f'{name}.pickle'
        super().__init__()

    def schedule(self) -> Self:
        """Apply the task."""
        all_tasks.add(self)
        return self

    def run_every(self, hours: int = 0, minutes: int = 0, seconds: int = 0) -> Self:
        """Schedule the task to run every so many seconds."""
        self._run_every_secs = timedelta(hours=hours, minutes=minutes, seconds=seconds).total_seconds()
        return self

    def run_once(self) -> Self:
        self._run_every_secs = None
        return self

    def in_total(self, times: int) -> Self:
        """Schedule the task to run a total of so many times."""
        if times <= 0:
            msg = 'times must be positive'
            raise ValueError(msg)
        self._in_total_times = times
        return self

    @abstractmethod
    def get_notion_objects(self) -> list[Any]:
        """Get all Notion objects to sync."""
        raise NotImplementedError()

    @abstractmethod
    def get_other_objects(self) -> list[Any]:
        """Get all other objects to sync."""
        raise NotImplementedError()

    @abstractmethod
    def notion_timestamp(self, obj: Any) -> datetime:
        """Get the timestamp of the Notion object."""
        raise NotImplementedError()

    @abstractmethod
    def other_timestamp(self, obj: Any) -> datetime:
        """Get the timestamp of the other object."""
        raise NotImplementedError()

    @abstractmethod
    def notion_id(self, obj: Any) -> ID:
        """Get the id of the Notion object."""
        raise NotImplementedError()

    @abstractmethod
    def other_id(self, obj: Any) -> ID:
        """Get the id of the other object."""
        raise NotImplementedError()

    @abstractmethod
    def notion_hash(self, obj: Any) -> str:
        """Get the hash of the Notion object for object mapping/linking."""
        raise NotImplementedError()

    @abstractmethod
    def other_hash(self, obj: Any) -> str:
        """Get the hash of the other object for object mapping/linking."""
        raise NotImplementedError()

    @abstractmethod
    def notion_to_dict(self, obj: Any) -> dict[str, Any]:
        """Convert a Notion object to a dictionary."""
        raise NotImplementedError()

    @abstractmethod
    def other_to_dict(self, obj: Any) -> dict[str, Any]:
        """Convert another object to a dictionary."""
        raise NotImplementedError()

    @abstractmethod
    def notion_update_obj(self, obj: Any, attr: str, value: Any) -> None:
        """Set an attribute of the Notion object, e.g. page."""
        raise NotImplementedError()

    @abstractmethod
    def other_update_obj(self, obj: Any, attr: str, value: Any) -> None:
        """Set an attribute of the other object."""
        raise NotImplementedError()

    @abstractmethod
    def notion_delete_obj(self, obj: Any) -> None:
        """Delete the page."""
        raise NotImplementedError()

    @abstractmethod
    def other_delete_obj(self, obj: Any) -> None:
        """Delete the other object."""
        raise NotImplementedError()

    @abstractmethod
    def notion_create_obj(self, **kwargs: Any) -> Any:
        """Create a new page."""
        raise NotImplementedError()

    @abstractmethod
    def other_create_obj(self, **kwargs: Any) -> Any:
        """Create a new other object."""
        raise NotImplementedError()

    def sync_notion_deleted(self, state: State, notion_objs: dict[ID, Any], other_objs: dict[ID, Any]) -> State:
        """Sync an object in the state that was deleted in Notion."""
        for notion_id, other_id in state.ids.copy().items():
            if notion_id not in notion_objs:
                if other_id in other_objs:
                    _logger.debug(f'Deleting other object with id {other_id}')
                    self.other_delete_obj(other_objs[other_id])
                    del other_objs[other_id]
                del state.ids[notion_id]
                del state.objs[notion_id]
        return state

    def sync_other_deleted(self, state: State, notion_objs: dict[ID, Any], other_objs: dict[ID, Any]) -> State:
        """Sync an object in the state that was deleted in the other service."""
        for notion_id, other_id in state.ids.copy().items():
            if other_id not in other_objs:
                if notion_id in notion_objs:
                    _logger.debug(f'Deleting Notion object with id {notion_id}')
                    self.notion_delete_obj(notion_objs[notion_id])
                    del notion_objs[notion_id]
                del state.ids[notion_id]
                del state.objs[notion_id]
        return state

    def sync_notion_created(self, state: State, notion_objs: dict[ID, Any]) -> State:
        """Sync an object not in the state and created in Notion."""
        for notion_id, notion_obj in notion_objs.items():
            if notion_id not in state.objs:
                notion_obj_dct = self.notion_to_dict(notion_obj)
                other_obj_dct = {
                    other_attr: notion_obj_dct[notion_attr] for notion_attr, other_attr in self.attr_map.items()
                }
                _logger.debug(f'Creating other object with attributes: {other_obj_dct}')
                other_obj = self.other_create_obj(**other_obj_dct)

                if self.other_to_dict(other_obj) != other_obj_dct:
                    msg = 'The other object created by Notion does not match the expected object.'
                    raise RuntimeError(msg)

                other_id = self.other_id(other_obj)
                state.ids[notion_id] = other_id
                state.objs[notion_id] = notion_obj_dct
        return state

    def sync_other_created(self, state: State, other_objs: dict[ID, Any]) -> State:
        """Sync an object not in the state and created in other service."""
        for other_id, other_obj in other_objs.items():
            if other_id not in state.ids.values():
                other_obj_dct = self.other_to_dict(other_obj)
                notion_obj_dct = {
                    notion_attr: other_obj_dct[other_attr] for notion_attr, other_attr in self.attr_map.items()
                }
                _logger.debug(f'Creating Notion object with attributes: {notion_obj_dct}')
                notion_obj = self.notion_create_obj(**notion_obj_dct)

                if self.notion_to_dict(notion_obj) != notion_obj_dct:
                    msg = 'The Notion object created by other service does not match the expected object.'
                    raise RuntimeError(msg)

                notion_id = self.notion_id(notion_obj)
                state.ids[notion_id] = other_id
                state.objs[notion_id] = notion_obj_dct
        return state

    def resolve_conflict(self, notion_obj: Any, other_obj: Any, notion_attr: str, other_attr: str) -> Any:
        """Resolve a conflict between two objects on an attribute."""
        _logger.debug(
            f'Resolving conflict on attribute {notion_attr} of Notion object {self.notion_id(notion_obj)}'
            f'and attribute {other_attr} of other object {self.other_id(other_obj)}.'
        )
        notion_obj_dct, other_obj_dct = self.notion_to_dict(notion_obj), self.other_to_dict(other_obj)

        if self.conflict_mode == ConflictMode.NOTION:
            _logger.debug(f'Updating other object with attribute {other_attr} to {notion_obj_dct[notion_attr]}')
            self.other_update_obj(other_obj, other_attr, notion_obj_dct[notion_attr])
            return notion_obj_dct[notion_attr]
        elif self.conflict_mode == ConflictMode.OTHER:
            _logger.debug(f'Updating Notion object with attribute {notion_attr} to {other_obj_dct[other_attr]}')
            self.notion_update_obj(notion_obj, notion_attr, other_obj_dct[other_attr])
            return other_obj_dct[other_attr]
        elif self.conflict_mode == ConflictMode.NEWER:
            if self.notion_timestamp(notion_obj) > self.other_timestamp(other_obj):
                _logger.debug(f'Updating other object with attribute {other_attr} to {notion_obj_dct[notion_attr]}')
                self.other_update_obj(other_obj, other_attr, notion_obj_dct[notion_attr])
                return notion_obj_dct[notion_attr]
            else:
                _logger.debug(f'Updating Notion object with attribute {notion_attr} to {other_obj_dct[other_attr]}')
                self.notion_update_obj(notion_obj, notion_attr, other_obj_dct[other_attr])
                return other_obj_dct[other_attr]
        else:
            msg = f'Unknown conflict mode {self.conflict_mode}'
            raise RuntimeError(msg)

    def initial_sync(self, notion_objs: dict[ID, Any], other_objs: dict[ID, Any]) -> State:
        """Make the initial state.

        This is a two-way sync, i.e. the objects are compared and the differences are resolved.
        """
        _logger.debug('Performing initial sync...')
        state = State()
        notion_hashes = {self.notion_hash(obj): obj_id for obj_id, obj in notion_objs.items()}
        other_hashes = {self.other_hash(obj): obj_id for obj_id, obj in other_objs.items()}

        for obj_hash in notion_hashes.keys() & other_hashes.keys():
            notion_id, other_id = notion_hashes[obj_hash], other_hashes[obj_hash]

            state.ids[notion_id] = other_id
            state_obj = state.objs[notion_id] = {}

            notion_obj, other_obj = notion_objs[notion_id], other_objs[other_id]
            notion_obj_dct, other_obj_dct = self.notion_to_dict(notion_obj), self.other_to_dict(other_obj)

            for notion_attr, other_attr in self.attr_map.items():
                if notion_obj_dct[notion_attr] == other_obj_dct[other_attr]:
                    state_obj[notion_attr] = notion_obj_dct[notion_attr]
                else:
                    state_obj[notion_attr] = self.resolve_conflict(notion_obj, other_obj, notion_attr, other_attr)

        return state

    def sync_state_changes(self, state: State, notion_objs: dict[ID, Any], other_objs: dict[ID, Any]) -> State:
        """Sync changes with respect to the state and update the state.

        This is a three-way sync, i.e. the objects are compared to the state as base and the differences are resolved.
        """
        _logger.debug('Performing state sync...')
        for notion_id, state_obj_dct in state.objs.items():
            notion_obj, other_obj = notion_objs[notion_id], other_objs[state.ids[notion_id]]
            notion_obj_dct, other_obj_dct = self.notion_to_dict(notion_obj), self.other_to_dict(other_obj)

            for notion_attr, other_attr in self.attr_map.items():
                if notion_obj_dct[notion_attr] != state_obj_dct[notion_attr] == other_obj_dct[other_attr]:
                    _logger.debug(f'Updating other object with attribute {other_attr} to {notion_obj_dct[notion_attr]}')
                    self.other_update_obj(other_obj, other_attr, notion_obj_dct[notion_attr])
                    state_obj_dct[notion_attr] = notion_obj_dct[notion_attr]
                elif notion_obj_dct[notion_attr] == state_obj_dct[notion_attr] != other_obj_dct[other_attr]:
                    _logger.debug(f'Updating Notion object with attribute {notion_attr} to {other_obj_dct[other_attr]}')
                    self.notion_update_obj(notion_obj, notion_attr, other_obj_dct[other_attr])
                    state_obj_dct[notion_attr] = notion_obj_dct[notion_attr]
                elif notion_obj_dct[notion_attr] != state_obj_dct[notion_attr] != other_obj_dct[other_attr]:
                    state_obj_dct[notion_attr] = self.resolve_conflict(notion_obj, other_obj, notion_attr, other_attr)

        return state

    def sync(self, state: State | None) -> State:
        """The actual sync operation.

        The state holds the synced objects and their attributes as a dictionary of Notion attributes.
        """
        notion_objs = {self.notion_id(obj): obj for obj in self.get_notion_objects()}
        other_objs = {self.other_id(obj): obj for obj in self.get_other_objects()}

        if state is None:
            state = self.initial_sync(notion_objs, other_objs)
        else:
            state = self.sync_notion_deleted(state, notion_objs, other_objs)
            state = self.sync_other_deleted(state, notion_objs, other_objs)
            state = self.sync_state_changes(state, notion_objs, other_objs)

        state = self.sync_notion_created(state, notion_objs)
        state = self.sync_other_created(state, other_objs)

        return state

    def __await__(self) -> Generator[Any, None, None]:
        """Delegate the await to the __call__ method"""
        return self().__await__()

    async def __call__(self) -> None:
        """Run the task as scheduled."""
        self.state_path.parent.mkdir(parents=True, exist_ok=True)
        state = pickle.loads(self.state_path.read_bytes()) if self.state_path.exists() else None  # noqa: S301

        while True:
            state = self.sync(state)
            self.state_path.write_bytes(pickle.dumps(state))

            if self._in_total_times is not None:
                self._in_total_times -= 1
                if self._in_total_times <= 0:
                    break

            if self._run_every_secs is not None:
                await asyncio.sleep(self._run_every_secs)

            if self._run_every_secs is None and self._in_total_times is None:
                break


all_tasks: set[SyncTask] = set()
"""All tasks that have been created so far."""


def run_all_tasks(*, debug: bool | None = None) -> None:
    """Run all scheduled tasks."""

    async def gather_all_tasks() -> list[Any]:
        return await asyncio.gather(*all_tasks)

    _logger.info('Running all scheduled tasks...')
    asyncio.run(gather_all_tasks(), debug=debug)



"""Low-level object-oriented API for Notion.

This subpackage provides an object-oriented API for Notion by wrapping the JSON-structured
requests and responses of [notion-sdk-py] into Python objects using pydantic. Also paginated
result lists by the API are resolved automatically.

The code was taken originally from [Notional] by Jason Heddings and is MIT-licensed.
Due to the Pydantic v2 migration and several other design changes, it was heavily refactored since then.

[Notional]: https://github.com/jheddings/notional
[notion-sdk-py]: https://github.com/ramnes/notion-sdk-py/
"""

from __future__ import annotations

import logging
import platform
from importlib.metadata import version
from typing import TYPE_CHECKING, Any

import httpx
import notion_client

from ultimate_notion import __version__
from ultimate_notion.config import get_cfg_file
from ultimate_notion.obj_api.endpoints import NotionAPI

if TYPE_CHECKING:
    from ultimate_notion.config import Config


_logger = logging.getLogger(__name__)


def _get_default_user_agent() -> str:
    """Return the default user agent string for the Notion client."""
    python_version = platform.python_version()
    os_name = platform.system()
    architecture = platform.machine()
    httpx_version = httpx.__version__
    notion_sdk_version = version('notion-client')
    return ' '.join(
        f'ultimate-notion/{__version__} (https://ultimate-notion.com/)'
        f'python/{python_version}'
        f'{os_name}/{architecture}'
        f'notion-sdk-py/{notion_sdk_version}'
        f'httpx/{httpx_version}'
    )


def create_notion_client(cfg: Config, **kwargs: Any) -> notion_client.Client:
    """Create a Notion client with the given authentication token."""
    if (auth := cfg.ultimate_notion.token) is None:
        msg = f'No Notion token found! Check {get_cfg_file()}.'
        raise RuntimeError(msg)

    # Set same sane defaults as notion_client defines its own logger
    kwargs.setdefault('logger', logging.getLogger('notion_client'))
    kwargs.setdefault('log_level', logging.NOTSET)

    def log_request(request: httpx.Request) -> None:
        msg = f'Request: {request.method} {request.url}'
        try:
            if request.content:
                msg += f'\n{request.content.decode("utf-8") if isinstance(request.content, bytes) else request.content}'
        except httpx.RequestNotRead:
            # For streaming requests (like file uploads), we can't access content without reading it first
            msg += '\n<streaming content>'
        _logger.debug(msg)

    def log_response(response: httpx.Response) -> None:
        msg = f'Response: {response.status_code} {response.url}'
        response.read()  # Ensure that the response content is fully loaded. Memory schouldn't be an issue here.
        if response.content:
            msg += f'\n{response.content.decode("utf-8") if isinstance(response.content, bytes) else response.content}'
        _logger.debug(msg)

    user_agent = kwargs.pop('user_agent', _get_default_user_agent())
    httpx_client = httpx.Client(event_hooks={'request': [log_request], 'response': [log_response]})
    client = notion_client.Client(auth=auth, client=httpx_client, **kwargs)
    # we need to set the user agent manually, because notion_client ovewrites it during initialization
    httpx_client.headers['User-Agent'] = user_agent
    return client


__all__ = ['NotionAPI', 'create_notion_client']



"""Additional utilities that fit nowhere else."""

from __future__ import annotations

import datetime as dt
import json
import re
import textwrap
import types
from collections.abc import Callable, Generator, Iterator, Mapping, Sequence
from contextlib import contextmanager
from copy import deepcopy
from functools import update_wrapper
from hashlib import sha256
from itertools import chain
from pathlib import Path
from typing import Any, Generic, ParamSpec, TypeAlias, TypeVar

import numpy as np
import pendulum as pnd
import tomli_w
from numpy.typing import NDArray
from packaging.version import Version
from pydantic import BaseModel
from typing_extensions import Self

from ultimate_notion import __version__
from ultimate_notion.errors import EmptyListError, MultipleItemsError

T = TypeVar('T')  # ToDo: Use new syntax when requires-python >= 3.12
P = ParamSpec('P')  # ToDo: Use new syntax when requires-python >= 3.12


class SList(list[T]):
    """A list that holds often only a single element."""

    def item(self) -> T:
        if len(self) == 1:
            return self[0]
        elif len(self) == 0:
            msg = 'list is empty'
            raise EmptyListError(msg)
        else:
            msg = 'list has multiple items'
            raise MultipleItemsError(msg)


def flatten(nested_list: Sequence[Sequence[T]], /) -> list[T]:
    """Flatten a nested list."""
    return list(chain.from_iterable(nested_list))


def safe_list_get(lst: Sequence[T], idx: int, *, default: T | None = None) -> T | None:
    """Get the element at the index of the list or return the default value."""
    try:
        return lst[idx]
    except IndexError:
        return default


def is_notebook() -> bool:
    """Determine if we are running within a Jupyter notebook."""
    try:
        from IPython import get_ipython  # noqa: PLC0415
    except ModuleNotFoundError:
        return False  # Probably standard Python interpreter
    else:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True  # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)


class StoredRetvalsFunctor(Generic[P, T]):
    """A decorator that stores the return values of a function for later use."""

    def __init__(self, func: Callable[P, T]) -> None:
        # Kopiert __name__, __doc__, __module__, __annotations__, __wrapped__
        update_wrapper(self, func)
        self._func = func
        self.retvals: list[T] = []

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        rv = self._func(*args, **kwargs)
        self.retvals.append(rv)
        return rv

    # Ensures that the decorator also binds correctly to methods (self/cls)
    def __get__(self, obj: Any, objtype: type | None = None) -> Self | types.MethodType:
        if obj is None:
            return self
        return types.MethodType(self, obj)  # bind 'obj' as first argument


def store_retvals(func: Callable[P, T]) -> StoredRetvalsFunctor[P, T]:
    """Decorator storing the return values as function attribute for later cleanups.

    This can be used for instance in a generator like this:
    ```
    @pytest.fixture
    def create_blank_db(notion, test_area):
        @store_retvals
        def nested_func(db_name):
            db = notion.databases.create(
                parent=test_area,
                title=db_name,
                schema={
                    'Name': schema.Title(),
                },
            )
            return db

        yield nested_func

        # clean up by deleting the db of each prior call
        for db in nested_func.retvals:
            notion.databases.delete(db)
    ```
    """
    return StoredRetvalsFunctor(func)


def find_indices(
    elements: NDArray[np.int_] | Sequence[Any], total_set: NDArray[np.int_] | Sequence[Any]
) -> NDArray[np.int_]:
    """Finds the indices of the elements in the total set."""
    if not isinstance(total_set, np.ndarray):
        total_set = np.array(total_set)
    mask = np.isin(total_set, elements)
    indices = np.where(mask)[0]
    lookup = dict(zip(total_set[mask], indices, strict=True))
    result = np.array([lookup.get(x) for x in elements])
    return result


def find_index(elem: Any, lst: list[Any]) -> int | None:
    """Find the index of the element in the list or return `None`."""
    try:
        return lst.index(elem)
    except ValueError:
        return None


def deepcopy_with_sharing(obj: T, shared_attributes: Sequence[str], memo: dict[int, Any] | None = None) -> T:
    """Like `deepcopy` but specified attributes are shared.

    Deepcopy an object, except for a given list of attributes, which should
    be shared between the original object and its copy.

    Args:
        obj: some object to copy
        shared_attributes: A list of strings identifying the attributes that should be shared instead of copied.
        memo: dictionary passed into __deepcopy__. Ignore this argument if not calling from within __deepcopy__.

    Example:
        ```python
        class A(object):
            def __init__(self):
                self.copy_me = []
                self.share_me = []

            def __deepcopy__(self, memo):
                return deepcopy_with_sharing(
                    self, shared_attribute_names=['share_me'], memo=memo
                )


        a = A()
        b = deepcopy(a)
        assert a.copy_me is not b.copy_me
        assert a.share_me is b.share_me

        c = deepcopy(b)
        assert c.copy_me is not b.copy_me
        assert c.share_me is b.share_me
        ```

    Original from https://stackoverflow.com/a/24621200
    """
    shared_attrs = {k: getattr(obj, k) for k in shared_attributes}

    if hasattr(obj, '__deepcopy__'):
        # Do hack to prevent infinite recursion in call to deepcopy
        deepcopy_method = obj.__deepcopy__
        obj.__deepcopy__ = None

    for attr in shared_attributes:
        del obj.__dict__[attr]

    clone = deepcopy(obj, memo)

    for attr, val in shared_attrs.items():
        setattr(obj, attr, val)
        setattr(clone, attr, val)

    if hasattr(obj, '__deepcopy__'):
        # Undo hack
        obj.__deepcopy__ = deepcopy_method
        del clone.__deepcopy__  # type: ignore[attr-defined]

    return clone


KT = TypeVar('KT')  # ToDo: Use new syntax when requires-python >= 3.12
VT = TypeVar('VT')  # ToDo: Use new syntax when requires-python >= 3.12


def dict_diff(dct1: Mapping[KT, VT], dct2: Mapping[KT, VT]) -> tuple[list[KT], list[KT], dict[KT, tuple[VT, VT]]]:
    """Returns the added keys, removed keys and keys of changed values of both dictionaries."""
    set1, set2 = set(dct1.keys()), set(dct2.keys())
    keys_added = list(set2 - set1)
    keys_removed = list(set1 - set2)
    values_changed = {key: (dct1[key], dct2[key]) for key in set1 & set2 if dct1[key] != dct2[key]}
    return keys_added, keys_removed, values_changed


def dict_diff_str(dct1: Mapping[KT, VT], dct2: Mapping[KT, VT]) -> tuple[list[str], list[str], list[str]]:
    """Returns the added keys, removed keys and keys of changed values of both dictionaries."""
    keys_added, keys_removed, values_changed = dict_diff(dct1, dct2)
    keys_added_str = [str(k) for k in keys_added]
    keys_removed_str = [str(k) for k in keys_removed]
    keys_changed_str = [f'{k}: {v[0]} -> {v[1]}' for k, v in values_changed.items()]
    return keys_added_str, keys_removed_str, keys_changed_str


def convert_md_to_py(path: Path | str, *, target_path: Path | str | None = None) -> None:
    """Converts a Markdown file to a py file by extracting all python codeblocks

    Args:
        path: Path to the Markdown file to convert
        target_path: Path to save the new Python file. If not provided, the new file will be the same file with .py

    !!! warning

        If a file with the same name already exists, it will be overwritten.
    """
    if isinstance(path, str):
        path = Path(path)
    if not path.is_file():
        msg = f'{path} is no file!'
        raise RuntimeError(msg)

    if target_path is None:
        target_path = path.with_suffix('.py')
    elif isinstance(target_path, str):
        target_path = Path(target_path)

    md_str = path.read_text()

    def check_codeblock(block: str) -> str:
        first_line = block.split('\n', maxsplit=1)[0]
        if first_line[3:] != 'python':
            return ''
        return '\n'.join(block.split('\n')[1:])

    docstring = textwrap.dedent(md_str)
    in_block = False
    block = ''
    codeblocks = []
    for line in docstring.split('\n'):
        if line.startswith('```'):
            if in_block:
                codeblocks.append(check_codeblock(block))
                block = ''
            in_block = not in_block
        if in_block:
            block += line + '\n'
    py_str = '\n'.join([c for c in codeblocks if c != ''])

    target_path.with_suffix('.py').write_text(py_str)


def str_hash(*args: str, n_chars: int = 16) -> str:
    """Hashes string arguments to a n-character string."""
    return sha256(''.join(args).encode('utf-8')).hexdigest()[:n_chars]


def rank(arr: NDArray[np.int_]) -> NDArray[np.int_]:
    """Returns the rank of the elements in the array and gives the same rank to equal elements."""
    mask = np.argsort(arr)
    rank = np.zeros_like(arr)
    rank[1:] = np.cumsum(np.where(np.diff(arr[mask]) != 0, 1, 0))
    return rank[np.argsort(mask)]


def is_stable_version(version_str: str) -> bool:
    """Return whether the given version is a stable release."""
    version = Version(version_str)
    return not (version.is_prerelease or version.is_devrelease or version.is_postrelease)


def is_stable_release() -> bool:
    """Return whether the current version is a stable release."""
    return is_stable_version(__version__)


DateTimeOrRange: TypeAlias = dt.datetime | dt.date | pnd.Interval[pnd.DateTime] | pnd.Interval[pnd.Date]
"""A type alias for various date, date time and interval representations."""


def parse_dt_str(dt_str: str) -> DateTimeOrRange:
    """Parse typical Notion date/datetime/interval strings to pendulum objects.

    If no timezone is provided assume local timezone and convert everything else to UTC for consistency."""

    def set_tz(dt_spec: pnd.DateTime | pnd.Date | dt.datetime | dt.date) -> pnd.DateTime | pnd.Date:
        """Set the timezone of the datetime specifier object if necessary."""
        match dt_spec:
            case pnd.DateTime() if dt_spec.tz is None:
                return dt_spec.in_tz('local')
            case pnd.DateTime():
                return dt_spec.in_tz('UTC')  # to avoid unnamed timezones we convert to UTC
            case pnd.Date():
                return dt_spec  # as it is a date and has no tz information
            case _:
                msg = f'Unexpected type `{type(dt_spec)}` for `{dt_spec}`'
                raise TypeError(msg)

    # Handle strings with "Europe/Berlin" and "UTC" style timezone
    if match := re.match(r'(.+)\s+([A-Za-z/]+)$', dt_str.strip()):
        dt_part, tz_part = match.groups()
        dt_spec = pnd.parse(dt_part, exact=True, tz=None)
        match dt_spec:
            case pnd.DateTime():
                dt_spec = dt_spec.in_tz(tz_part)
            case _:
                msg = f'Expected a datetime string but got {dt_str}'
                raise ValueError(msg)
    else:
        dt_spec = pnd.parse(dt_str, exact=True, tz=None)

    match dt_spec:
        case pnd.DateTime():
            return set_tz(dt_spec)
        case pnd.Date():
            return dt_spec
        case pnd.Interval():
            # We extend the interval to the full day if only a date is given
            start, end = set_tz(dt_spec.start), set_tz(dt_spec.end)
            if not isinstance(dt_spec.start, pnd.DateTime):
                start = pnd.datetime(start.year, start.month, start.day, 0, 0, 0)
            if not isinstance(dt_spec.end, pnd.DateTime):
                end = pnd.datetime(end.year, end.month, end.day, 23, 59, 59)
            return pnd.Interval(start=start, end=end)
        case _:
            msg = f'Unexpected parsing result of type {type(dt_spec)} for {dt_str}'
            raise TypeError(msg)


def is_dt_str(dt_str: str) -> bool:
    """Check if the given string is a valid datetime string."""
    try:
        parse_dt_str(dt_str)
        return True
    except (ValueError, TypeError):
        return False


def to_pendulum(dt_spec: str | DateTimeOrRange) -> DateTimeOrRange:
    """Convert a datetime or date object to a pendulum object."""
    match dt_spec:
        case pnd.DateTime() | pnd.Date() | pnd.Interval():
            return dt_spec
        case str():
            return parse_dt_str(dt_spec)
        case dt.datetime() if dt_spec.tzinfo is None:
            return pnd.instance(dt_spec, tz='local')
        case dt.datetime():
            return pnd.instance(dt_spec).in_tz('UTC')  # to avoid unnamed timezones we convert to UTC
        case dt.date():
            return pnd.instance(dt_spec)
        case _:
            msg = f'Unexpected type {type(dt_spec)} for {dt_spec}'
            raise TypeError(msg)


@contextmanager
def temp_timezone(tz: str | pnd.Timezone) -> Iterator[None]:
    """Temporarily set the local timezone to the given timezone. Mostly used by unit tests."""
    if not isinstance(tz, pnd.Timezone):
        tz = pnd.timezone(tz)

    current_tz = pnd.local_timezone()
    if not isinstance(current_tz, pnd.Timezone):
        msg = f'Expected a Timezone object but got type {type(current_tz)}.'
        raise RuntimeError(msg)
    pnd.set_local_timezone(tz)
    try:
        yield
    finally:
        pnd.set_local_timezone(current_tz)


PT = TypeVar('PT', bound=BaseModel)  # ToDo: Use new syntax when requires-python >= 3.12


def set_attr_none(
    obj: PT, attr_paths: str | Sequence[str] | None, *, inplace: bool = False, missing_ok: bool = False
) -> PT:
    """Set the attributes given by a potentially nested path to None.

    `None` attributes will be removed during serialization by default.
    """
    if attr_paths is None:
        return obj
    if isinstance(attr_paths, str):
        attr_paths = [attr_paths]

    if not inplace:
        obj = obj.model_copy(deep=True)
    for attr_path in attr_paths:
        attrs = attr_path.split('.')

        curr_obj: Any = obj
        for lvl, attr in enumerate(attrs[:-1]):
            curr_obj = getattr(curr_obj, attr, None)
            if curr_obj is None and not missing_ok:
                msg = f'{attr} does not exist in {".".join(attrs[: lvl - 1]) if lvl > 1 else "the object"}.'
                raise AttributeError(msg)

        last_attr = attrs[-1]
        if hasattr(curr_obj, last_attr):
            setattr(curr_obj, last_attr, None)
        elif not missing_ok:
            msg = f'{last_attr} does not exist in {".".join(attrs[:-2]) if len(attrs) > 1 else "the object"}.'
            raise AttributeError(msg)

    return obj


@contextmanager
def temp_attr(obj: object, **kwargs: Any) -> Generator[None, None, None]:
    """
    Temporarily sets multiple attributes of an object to specified values,
    and restores their original values after the context exits.

    Args:
        obj (object): The object whose attributes will be modified.
        **kwargs (Any): The attributes and their temporary values to modify.
    """
    orig_values = {attr: getattr(obj, attr, None) for attr in kwargs}
    for attr, new_value in kwargs.items():
        setattr(obj, attr, new_value)
    try:
        yield
    finally:
        for attr, original_value in orig_values.items():
            setattr(obj, attr, original_value)


def rec_apply(func: Callable[[Any], Any], obj: Any) -> Any:
    """
    Recursively applies a function `func` to all elements in a nested structure.

    - Applies `func` to every non-container element.
    - Recurses into lists, tuples and dicts.
    - Strings are treated as atomic elements and are **not** considered containers.

    Example:
        rows = [[1, 2], [3, [4, 5]]]
        result = recursive_apply(rows, lambda x: x * 2)
        print(result)  # [[2, 4], [6, [8, 10]]]
    """
    if isinstance(obj, str):  # str is a sequence!
        return func(obj)
    elif isinstance(obj, Sequence):
        return [rec_apply(func, item) for item in obj]
    elif isinstance(obj, Mapping):
        return {k: rec_apply(func, v) for k, v in obj.items()}
    else:
        return func(obj)


def pydantic_to_toml(model: BaseModel) -> str:
    """Convert a Pydantic model to a TOML string."""
    json_dct = json.loads(model.model_dump_json())
    # Remove None as TOML doesn't support null/None values.
    json_dct = rec_apply(lambda x: '' if x is None else x, json_dct)
    return tomli_w.dumps(json_dct)


def pydantic_apply(obj: PT, func: Callable[[Any, Any], Any]) -> PT:
    """Apply a function to all fields, i.e. (name, value), of a Pydantic model recursively.

    The transformed model is returned as a copy, leaving the original model unchanged.
    """

    def _trans_list_elem(name: str, elem: Any) -> Any:
        if isinstance(elem, BaseModel):
            return pydantic_apply(elem, func)
        elif isinstance(elem, list):  # happens for cells of a TableRow
            return [pydantic_apply(item, func) for item in elem]
        else:
            return func(name, elem)

    upd_obj = obj.model_copy(deep=True)
    for name in upd_obj.__class__.model_fields:
        value = getattr(upd_obj, name)
        if isinstance(value, BaseModel):  # First dive into nested BaseModels ...
            setattr(upd_obj, name, value := pydantic_apply(value, func))
        # ... then work on the key-value pairs directly, where := updates value to remove, e.g. empty lists
        setattr(upd_obj, name, value := func(name, value))

        if isinstance(value, list):
            new_list = [_trans_list_elem(name, item) for item in value]
            setattr(upd_obj, name, new_list)
        elif isinstance(value, dict):
            new_dict = {
                k: pydantic_apply(v, func) if isinstance(v, BaseModel) else func(name, v) for k, v in value.items()
            }
            setattr(upd_obj, name, new_dict)

    return upd_obj



"""Client for Google Tasks API.

Follow this quickstart guide to enable the API and create the necessary credentials:
https://developers.google.com/tasks/quickstart/python

Official Python API documentation: https://googleapis.github.io/google-api-python-client/docs/dyn/tasks_v1.html
Official REST documentation: https://developers.google.com/tasks/reference/rest

!!! danger
    The Google Task API does not support setting time information for due dates.
    Thus, the time information of a datetime object is truncated and a pure date is obtained.
    https://issuetracker.google.com/issues/128979662
"""

from __future__ import annotations

from collections.abc import Iterator
from datetime import date, datetime, time, timezone
from enum import Enum
from types import TracebackType
from typing import Any, Literal

from google.auth.exceptions import RefreshError
from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import Resource, build
from pydantic import BaseModel, ConfigDict, Field, HttpUrl

from ultimate_notion.config import Config, get_or_create_cfg
from ultimate_notion.utils import SList, is_stable_release

DEFAULT_LIST_ID_LEN = 32
"""Length of the ID of the Google default tasklist."""
MAX_RESULTS_PER_PAGE = 100
"""Maximum number of results per page when fetching all tasks."""


class Scope(str, Enum):
    TASKS_RO = 'https://www.googleapis.com/auth/tasks.readonly'
    """Allows read-only access to Google Tasks"""
    TASKS_RW = 'https://www.googleapis.com/auth/tasks'
    """Allows read/write access to Google Tasks"""


class Link(BaseModel):
    """Representation of a link in a Google Task."""

    model_config = ConfigDict(extra='ignore' if is_stable_release() else 'forbid')
    description: str
    Link: HttpUrl
    type: str


class Status(str, Enum):
    """Representation of the different statuses of a Google Task."""

    NEEDS_ACTION = 'needsAction'
    COMPLETED = 'completed'


class Kind(str, Enum):
    """Representation of the different kinds of Google Objects."""

    TASK = 'tasks#task'
    TASK_LIST = 'tasks#taskList'


class GObject(BaseModel):
    """Representation of a general Google Object from the Tasks API."""

    model_config = ConfigDict(extra='ignore' if is_stable_release() else 'forbid')

    id: str  # A003
    title_: str = Field(..., alias='title')
    kind: Kind
    etag: str
    updated: datetime
    self_link: HttpUrl = Field(..., alias='selfLink')
    _resource: Resource

    def __init__(self, resource: Resource, **data: Any) -> None:
        super().__init__(**data)
        self._resource = resource

    def __eq__(self, other: object) -> bool:
        if isinstance(other, GObject):
            return self.kind == self.kind and self.id == other.id
        else:
            return NotImplemented

    def __hash__(self) -> int:
        return hash(self.id)

    def _update(self, resp: dict[str, str]) -> None:
        """Updates this object with the given response from the API."""
        new_obj_dct = dict(resp, resource=self._resource)
        new_obj = self.model_validate(new_obj_dct)

        for k in new_obj.model_dump(by_alias=False):
            setattr(self, k, getattr(new_obj, k))


class GTask(GObject):
    """Representation of a Google Task."""

    kind: Literal[Kind.TASK]  # ToDo: This is nonsense, replace with proper generic
    completed_at: datetime | None = Field(alias='completed', default=None)
    is_deleted: bool = Field(alias='deleted', default=False)
    hidden: bool = False
    due_: datetime | None = Field(alias='due', default=None)
    notes_: str | None = Field(alias='notes', default=None)
    parent_id: str | None = Field(alias='parent', default=None)
    position_: str = Field(alias='position')
    status: Status = Status.NEEDS_ACTION
    links: list[Link] | None = None
    web_view_link: HttpUrl = Field(alias='webViewLink')

    @property
    def tasklist_id(self) -> str:
        """Returns the task list this task belongs to."""
        if self.self_link.path is None:
            msg = 'selfLink has no path to determine the id of the tasklist!'
            raise RuntimeError(msg)

        url_paths = self.self_link.path.split('/')
        if url_paths[3] != 'lists':
            msg = f'Unexpected URL path: {self.self_link.path}'
            raise RuntimeError(msg)
        return url_paths[4]

    @property
    def tasklist(self) -> GTaskList:
        """Returns the task list this task belongs to."""
        tasklist = self._resource.tasklists().get(tasklist=self.tasklist_id).execute()
        return GTaskList(resource=self._resource, **tasklist)

    def delete(self) -> GTask:
        """Deletes this task."""
        resource = self._resource.tasks()
        resource.delete(tasklist=self.tasklist_id, task=self.id).execute()
        resp = resource.get(tasklist=self.tasklist_id, task=self.id).execute()
        self._update(resp)
        return self

    @property
    def due(self) -> datetime | None:
        """Returns the due date of this task."""
        return self.due_

    @due.setter
    def due(self, due_date: datetime | date | None) -> None:
        """Sets the due date of this task."""
        due_date = assert_datetime(due_date)
        due_date_str = due_date if due_date is None else due_date.isoformat()
        resource = self._resource.tasks()
        resp = resource.patch(tasklist=self.tasklist_id, task=self.id, body={'due': due_date_str}).execute()
        self._update(resp)

    @property
    def notes(self) -> str | None:
        """Returns the notes of this task."""
        return self.notes_

    @notes.setter
    def notes(self, new_notes: str | None) -> None:
        """Sets the notes of this task."""
        resource = self._resource.tasks()
        resp = resource.patch(tasklist=self.tasklist_id, task=self.id, body={'notes': new_notes}).execute()
        self._update(resp)

    @property
    def title(self) -> str:
        """Returns the title of this task."""
        return self.title_

    @title.setter
    def title(self, new_title: str) -> None:
        """Sets the title of this task."""
        resource = self._resource.tasks()
        resp = resource.patch(tasklist=self.tasklist_id, task=self.id, body={'title': new_title}).execute()
        self._update(resp)

    @property
    def is_completed(self) -> bool:
        """Returns whether this task is completed."""
        return self.status == Status.COMPLETED

    @is_completed.setter
    def is_completed(self, completed: bool) -> None:
        """Sets the completed status of this task."""
        resource = self._resource.tasks()
        resp = resource.patch(
            tasklist=self.tasklist_id,
            task=self.id,
            body={'status': Status.COMPLETED if completed else Status.NEEDS_ACTION},
        ).execute()
        self._update(resp)

    @property
    def position(self) -> int:
        """Returns the position of this task."""
        return int(self.position_)

    def position_after(self, previous: GTask | None = None) -> GTask:
        """Moves this task to the behind the given task."""
        previous_id = None if previous is None else previous.id
        resource = self._resource.tasks()
        resp = resource.move(tasklist=self.tasklist_id, task=self.id, previous=previous_id).execute()
        self._update(resp)
        return self

    @property
    def parent(self) -> GTask | None:
        """Returns the parent task of this task."""
        if self.parent_id is None:
            return None
        resource = self._resource.tasks()
        resp = resource.get(tasklist=self.tasklist_id, task=self.parent_id).execute()
        return GTask(resource=self._resource, **resp)

    @parent.setter
    def parent(self, parent: GTask | None) -> None:
        """Sets the parent of this task."""
        parent_id = None if parent is None else parent.id

        resource = self._resource.tasks()
        resp = resource.move(tasklist=self.tasklist_id, task=self.id, parent=parent_id).execute()
        self._update(resp)

    @property
    def children(self) -> list[GTask]:
        """Returns the children of this task."""
        return [task for task in self.tasklist.all_tasks() if task.parent_id == self.id]

    def reload(self) -> GTask:
        """Reloads this task from the API."""
        resource = self._resource.tasks()
        resp = resource.get(tasklist=self.tasklist_id, task=self.id).execute()
        self._update(resp)
        return self


class GTaskList(GObject):
    """Representation of a Google Task List."""

    kind: Literal[Kind.TASK_LIST]

    def all_tasks(self, *, show_deleted: bool = False) -> list[GTask]:
        """Returns a list of all tasks, completed or not, in this task list."""
        resource = self._resource.tasks()
        page_token = None
        tasks = []

        while True:
            results = resource.list(
                tasklist=self.id,
                maxResults=MAX_RESULTS_PER_PAGE,
                pageToken=page_token,
                showCompleted=True,
                showHidden=True,
                showDeleted=show_deleted,
            ).execute()
            items = results.get('items', [])
            tasks.extend([GTask(resource=self._resource, **item) for item in items])

            page_token = results.get('nextPageToken')
            if page_token is None:
                break

        return tasks

    def __iter__(self) -> Iterator[GTask]:  # type: ignore[override]
        """Returns an iterator over all tasks in this task list."""
        yield from self.all_tasks()

    def __len__(self) -> int:
        """Return the number of tasks in this task list."""
        return len(self.all_tasks())

    @property
    def is_empty(self) -> bool:
        """Is this task list empty?"""
        return len(self) == 0

    def __bool__(self) -> bool:
        """Overwrite default behaviour."""
        msg = 'Use .is_empty instead of bool(task_list) to check if a task list is empty.'
        raise RuntimeError(msg)

    def get_task(self, task_id: str) -> GTask:
        """Returns the task with the given ID."""
        resource = self._resource.tasks()
        task = resource.get(tasklist=self.id, task=task_id).execute()
        return GTask(resource=self._resource, **task)

    def create_task(self, title: str, due: date | datetime | None = None) -> GTask:
        """Creates a new task."""
        due = assert_datetime(due)
        resource = self._resource.tasks()
        body = {'title': title}
        if due is not None:
            body['due'] = due.isoformat()
        task = resource.insert(tasklist=self.id, body=body).execute()
        return GTask(resource=self._resource, **task)

    @property
    def is_default(self) -> bool:
        """Is this the default task list?"""
        return len(self.id) == DEFAULT_LIST_ID_LEN

    def delete(self) -> None:
        """Deletes this task list."""
        if len(self.id) == DEFAULT_LIST_ID_LEN:
            msg = 'This is the default tasklist and thus cannot be deleted!'
            raise RuntimeError(msg)
        resource = self._resource.tasklists()
        resource.delete(tasklist=self.id).execute()
        # We return None as the object is deleted

    @property
    def title(self) -> str:
        """Returns the title of this task list."""
        return self.title_

    @title.setter
    def title(self, new_title: str) -> None:
        """Sets the title of this task list."""
        resource = self._resource.tasklists()
        resp = resource.patch(tasklist=self.id, body={'title': new_title}).execute()
        self._update(resp)

    def reload(self) -> GTaskList:
        """Reloads this task list from the API."""
        resource = self._resource.tasklists()
        resp = resource.get(tasklist=self.id).execute()
        self._update(resp)
        return self


class GTasksClient:
    """Google API to easily handle Google Tasks."""

    read_only: bool
    _scopes: list[str]
    _config: Config
    resource: Resource

    def __init__(self, config: Config | None = None, *, read_only: bool = False) -> None:
        self.read_only = read_only
        self._scopes = [Scope.TASKS_RO.value] if read_only else [Scope.TASKS_RW.value]
        if config is None:
            config = get_or_create_cfg()
        self._config = config
        self.resource = self._build_resource()

    def _build_resource(self) -> Resource:
        if self._config.google is None:
            msg = 'Configurtion has no `google` section!'
            raise RuntimeError(msg)
        if (secret_path := self._config.google.client_secret_json) is None:
            msg = 'You have to set google.client_secret_json in your config.toml!'
            raise RuntimeError(msg)
        if (token_path := self._config.google.token_json) is None:
            msg = 'You have to set google.token_json in your config.toml!'
            raise RuntimeError(msg)
        if not secret_path.exists():
            msg = f'File {secret_path} does not exist!'
            raise RuntimeError(msg)

        creds = Credentials.from_authorized_user_file(token_path, self._scopes) if token_path.exists() else None
        if not creds or not creds.valid:
            if creds and creds.expired and creds.refresh_token:
                try:
                    creds.refresh(Request())
                except RefreshError as e:
                    msg = f'Error refreshing token. Please delete {token_path} and try again!'
                    raise RuntimeError(msg) from e
            else:
                flow = InstalledAppFlow.from_client_secrets_file(secret_path, self._scopes)
                creds = flow.run_local_server(port=0)
            # Save the credentials for the next run
            with open(token_path, 'w', encoding='utf8') as token:
                token.write(creds.to_json())

        return build('tasks', 'v1', credentials=creds)

    def recreate_token(self) -> None:
        """Recreate the current token using the scopes given at initialization."""
        if (gconfig := self._config.google) and (token_json := gconfig.token_json):
            token_json.unlink(missing_ok=True)
        self.resource = self._build_resource()

    def all_tasklists(self, max_results: int | None = None) -> list[GTaskList]:
        """Returns a list of all task lists."""
        page_token = None
        tasklists = []

        while True:
            results = self.resource.tasklists().list(maxResults=max_results, pageToken=page_token).execute()
            items = results.get('items', [])
            tasklists.extend([GTaskList(resource=self.resource, **item) for item in items])

            page_token = results.get('nextPageToken')
            if not page_token:
                break

        return tasklists

    def get_tasklist(self, tasklist_id: str = '@default') -> GTaskList:
        """Returns the task list with the given ID.

        If no ID is given, the default task list is returned.
        """
        tasklist = self.resource.tasklists().get(tasklist=tasklist_id).execute()
        return GTaskList(resource=self.resource, **tasklist)

    def search_tasklist(self, title: str) -> SList[GTaskList]:
        """Returns the task list with the given title."""
        tasklists = self.all_tasklists()
        return SList(tasklist for tasklist in tasklists if tasklist.title == title)

    def create_tasklist(self, title: str) -> GTaskList:
        """Creates a new task list."""
        tasklist = self.resource.tasklists().insert(body={'title': title}).execute()
        return GTaskList(resource=self.resource, **tasklist)

    def get_or_create_tasklist(self, title: str) -> GTaskList:
        """Returns the task list with the given title or creates it if it doesn't exist yet."""
        tasklists = self.search_tasklist(title)
        if len(tasklists) == 0:
            return self.create_tasklist(title)
        else:
            return tasklists.item()

    def close(self) -> None:
        """Closes the client."""
        self.resource.close()

    def __enter__(self) -> GTasksClient:
        return self

    def __exit__(
        self, exc_type: type[BaseException] | None, exc_value: BaseException | None, traceback: TracebackType | None
    ) -> None:
        self.close()


def assert_datetime(dt: date | datetime | None) -> datetime | None:
    """Asserts that the given object is a datetime object or None."""
    match dt:
        case None:
            return None
        case datetime():
            return dt
        case date():
            return datetime.combine(dt, time(tzinfo=timezone.utc))
        case _:
            msg = f'Expected a datetime object or None, but got {type(dt)}!'
            raise TypeError(msg)



import enum
import logging
import sys
from pathlib import Path
from typing import Annotated
from uuid import UUID

import filetype
import typer

from ultimate_notion import Session, __version__
from ultimate_notion.blocks import PDF, Audio, File, Image, Video
from ultimate_notion.config import get_cfg, get_cfg_file
from ultimate_notion.errors import UnknownPageError
from ultimate_notion.page import Page
from ultimate_notion.utils import pydantic_to_toml

_logger = logging.getLogger(__name__)


class LogLevel(str, enum.Enum):
    CRITICAL = 'critical'
    ERROR = 'error'
    WARNING = 'warning'
    INFO = 'info'
    DEBUG = 'debug'


def setup_logging(log_level: LogLevel) -> None:
    """Setup basic logging"""
    log_format = '[%(asctime)s] %(levelname)s:%(name)s:%(message)s'
    numeric_level = getattr(logging, log_level.upper(), None)
    logging.basicConfig(level=numeric_level, stream=sys.stdout, format=log_format, datefmt='%Y-%m-%d %H:%M:%S')


app = typer.Typer(
    name=f'Ultimate Notion {__version__}',
    help='🚀 The ultimate Python client for Notion!',
)


@app.callback()
def main(log_level: Annotated[LogLevel, typer.Option(help='Log level')] = LogLevel.WARNING) -> None:
    """Shared options for all commands"""
    setup_logging(log_level)


@app.command()
def config() -> None:
    """Display the current configuration file path and contents"""
    cfg_file = get_cfg_file()
    cfg = get_cfg()

    typer.echo(f'Config file: {cfg_file}')
    typer.echo('Configuration:')
    typer.echo('-' * 40)
    toml_string = pydantic_to_toml(cfg)
    typer.echo(toml_string)


@app.command()
def info() -> None:
    """Display information about the Notion integration"""
    typer.echo(f'Python version: {sys.version}')
    typer.echo(f'Ultimate Notion version: {__version__}')
    with Session() as session:
        this_integration = session.whoami()

    typer.echo('Notion integration:')
    typer.echo(f'- name: {this_integration.name}')
    typer.echo(f'- id: {this_integration.id}')

    typer.echo('Workspace info:')
    if workspace_info := this_integration.workspace_info:
        for key, value in workspace_info.model_dump().items():
            typer.echo(f'- {key}: {value}')


def _is_uuid(value: str) -> bool:
    """Check if a string is a valid UUID."""
    try:
        UUID(value)
        return True
    except ValueError:
        return False


def _get_block_class_for_file(file_path: Path) -> type[File | Image | Video | PDF | Audio]:
    """Determine the appropriate block class based on file content and extension.

    Uses the filetype library to detect file type by examining magic bytes,
    falling back to extension-based detection if needed.
    """
    # First try to detect by examining file content
    kind = filetype.guess(str(file_path))
    if kind is not None:
        mime_type = kind.mime
        if mime_type.startswith('image/'):
            return Image
        elif mime_type.startswith('video/'):
            return Video
        elif mime_type.startswith('audio/'):
            return Audio
        elif mime_type == 'application/pdf':
            return PDF

    return File


def _find_page_by_name(session: Session, page_name: str) -> Page:
    """Find a page by name, ensuring it's unique."""
    pages = session.search_page(page_name, exact=True)

    if not pages:
        typer.echo(f"Error: No page found with name '{page_name}'", err=True)
        raise typer.Exit(1)

    if len(pages) > 1:
        typer.echo(f"Error: Multiple pages found with name '{page_name}'. Please use the UUID instead.", err=True)
        typer.echo('Found pages:')
        for page in pages:
            typer.echo(f'  - {page.title} (ID: {page.id})')
        raise typer.Exit(1)

    return pages.item()


@app.command()
def upload(
    file_name: Annotated[str, typer.Argument(help='Path to the file to upload')],
    notion_page: Annotated[str, typer.Argument(help='Page name or UUID to upload the file to')],
) -> None:
    """Upload a file to a Notion page and append it as a block.

    The file will be uploaded to Notion and appended to the specified page.
    The block type is automatically determined based on the file extension:
    - Images (.png, .jpg, .jpeg, .gif, .webp, etc.) → Image block
    - Videos (.mp4, .avi, .mov, .wmv, etc.) → Video block
    - PDFs (.pdf) → PDF block
    - All other files → File block

    The page can be specified either by name or by UUID. If specified by name,
    the name must be unique (exact match).
    """
    file_path = Path(file_name)

    # Check if file exists
    if not file_path.exists():
        typer.echo(f"Error: File '{file_name}' does not exist", err=True)
        raise typer.Exit(1)

    if not file_path.is_file():
        typer.echo(f"Error: '{file_name}' is not a file", err=True)
        raise typer.Exit(1)

    with Session() as session:
        # Determine if notion_page is a UUID or name
        if _is_uuid(notion_page):
            try:
                page = session.get_page(notion_page)
                _logger.info(f"Found page by UUID: '{page.title}' (ID: {page.id})")
            except UnknownPageError as err:
                typer.echo(f"Error: Page with UUID '{notion_page}' not found", err=True)
                raise typer.Exit(1) from err
        else:
            _logger.info(f"Searching for page with name: '{notion_page}'")
            page = _find_page_by_name(session, notion_page)
            _logger.info(f"Found page: '{page.title}' (ID: {page.id})")

        _logger.info(f'Uploading file: {file_path.name}')
        with open(file_path, 'rb') as f:
            uploaded_file = session.upload(f, file_name=file_path.name)

        block_class = _get_block_class_for_file(file_path)
        block = block_class(uploaded_file)
        page.append(block)
        _logger.info(f"Successfully appended {block_class.__name__} block to page '{page.title}'")



"""Iterator classes for working with paginated API responses."""

from __future__ import annotations

import logging
from collections.abc import Awaitable, Callable, Iterator
from typing import Annotated, Any, Generic, TypeVar, cast

from pydantic import ConfigDict, Field
from pydantic.functional_validators import BeforeValidator

from ultimate_notion.obj_api.blocks import Block, Database, Page
from ultimate_notion.obj_api.core import GenericObject, NotionObject, TypedObject
from ultimate_notion.obj_api.objects import Comment, FileUpload, User
from ultimate_notion.obj_api.props import PropertyItem

MAX_PAGE_SIZE = 100

_logger = logging.getLogger(__name__)


def convert_to_notion_obj(
    data: dict[str, Any],
) -> Block | Page | Database | PropertyItem | User | GenericObject | FileUpload:
    """Convert a dictionary to the corresponding subtype of Notion Object.

    Used in the ObjectList below the convert the results from the Notion API.
    """
    obj_field = 'object'

    if obj_field not in data:
        msg = 'Unknown object in results'
        raise ValueError(msg)

    models: tuple[type[NotionObject], ...] = (Block, Page, Database, PropertyItem, User, Comment, FileUpload)
    model_mapping = {model.model_fields[obj_field].default: model for model in models}
    model_class = model_mapping.get(data[obj_field], GenericObject)
    return model_class.model_validate(data)


class ObjectList(NotionObject, TypedObject, object='list', polymorphic_base=True):
    """A paginated list of objects returned by the Notion API.

    More details in the [Notion API](https://developers.notion.com/reference/intro#responses).
    """

    results: list[Annotated[NotionObject, BeforeValidator(convert_to_notion_obj)]] = Field(default_factory=list)
    has_more: bool = False
    next_cursor: str | None = None


class BlockList(ObjectList, type='block'):
    """A list of Block objects returned by the Notion API."""

    class TypeData(GenericObject): ...

    block: TypeData


class PageOrDatabaseList(ObjectList, type='page_or_database'):
    """A list of Page or Database objects returned by the Notion API."""

    class TypeData(GenericObject): ...

    page_or_database: TypeData


class UserList(ObjectList, type='user'):
    """A list of User objects returned by the Notion API."""

    class TypeData(GenericObject): ...

    user: TypeData


class CommentList(ObjectList, type='comment'):
    """A list of Comment objects returned by the Notion API."""

    class TypeData(GenericObject): ...

    comment: TypeData


class PropertyItemList(ObjectList, type='property_item'):
    """A paginated list of property items returned by the Notion API.

    Property item lists contain one or more pages of basic property items.  These types
    do not typically match the schema for corresponding property values.
    """

    class TypeData(GenericObject):
        model_config = ConfigDict(extra='allow')  # for additional `type` field
        id: str
        type: str
        next_url: str | None  # not clear what this is for

    property_item: TypeData


class FileUploadList(ObjectList, type='file_upload'):
    """A list of FileUpload objects returned by the Notion API."""

    class TypeData(GenericObject): ...

    file_upload: TypeData


T = TypeVar('T', bound=NotionObject)


class EndpointIterator(Generic[T]):
    """Functor to iterate over results from a potentially paginated API response.

    In most cases `notion_obj` should be `ObjectList`. For some endpoints, like `PropertiesEndpoint`, and
    endpoint returns a list of property items or s single property item. In this case, `model_validate` should be
    `TypeAdapter(PropertyItemList | PropertyItem).validate_python` or whatever you expect to be returned.
    """

    has_more: bool | None = None
    page_num: int = -1
    total_items: int = -1
    next_cursor: str | None = None

    def __init__(
        self,
        endpoint: Callable[..., Any | Awaitable[Any]],
        *,
        pagination: Callable[..., Any | Awaitable[Any]] | None = None,
        model_validate: Callable[[Any], NotionObject] = ObjectList.model_validate,
    ):
        """Initialize an object list iterator for the specified endpoint."""
        self._endpoint = endpoint
        self._pagination = pagination
        self._model_validate = model_validate

    def __call__(self, **kwargs: Any) -> Iterator[T]:
        """Return a generator for this endpoint using the given parameters."""

        self.has_more = True
        self.page_num = 0
        self.total_items = 0

        if 'page_size' not in kwargs:
            kwargs['page_size'] = MAX_PAGE_SIZE

        self.next_cursor = kwargs.pop('start_cursor', None)

        while self.has_more:
            self.page_num += 1

            if self.next_cursor is None:
                msg = f'Fetching last page {self.page_num} of endpoint.'
            else:
                msg = f'Fetching page {self.page_num} of endpoint with next cursor {self.next_cursor}.'
            _logger.debug(msg)

            result_page = self._endpoint(start_cursor=self.next_cursor, **kwargs)
            obj_or_list = self._model_validate(result_page)

            if isinstance(obj_or_list, ObjectList):
                for obj in obj_or_list.results:
                    self.total_items += 1
                    yield cast(T, obj)

                self.next_cursor = obj_or_list.next_cursor
                self.has_more = obj_or_list.has_more and self.next_cursor is not None
                if self.has_more and self._pagination is not None:
                    # Some endpoints like `blocks.children.append` use a different endpoint for pagination.
                    # If not used, the action, e.g. appending blocks, will be repeated.
                    # You gotta love the Notion API...
                    self._endpoint = self._pagination
            else:
                yield cast(T, obj_or_list)
                self.has_more = False

            _logger.debug(f'Fetched {self.total_items} item(s) in total over {self.page_num} page(s).')



"""Query the database for pages."""

from __future__ import annotations

import datetime as dt
import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, cast

from pydantic import BaseModel, Field
from typing_extensions import Self

from ultimate_notion import props, schema
from ultimate_notion.core import get_active_session
from ultimate_notion.errors import EmptyDBError, FilterQueryError
from ultimate_notion.obj_api import query as obj_query
from ultimate_notion.obj_api.core import raise_unset
from ultimate_notion.obj_api.enums import ArrayQuantifier, FormulaType, RollupType, SortDirection
from ultimate_notion.option import Option
from ultimate_notion.page import Page
from ultimate_notion.user import User
from ultimate_notion.view import View

if TYPE_CHECKING:
    from ultimate_notion.database import Database
    from ultimate_notion.schema import Property


_logger = logging.getLogger(__name__)


class PageProperty(BaseModel):
    """Represents a property of a page.

    !!! note

        We override some magic methods to allow for more natural query building in an unorthodox way.
        Be aware that for instance the comparison operator == will not return boolean values but
        instances of the corresponding condition classes.
    """

    name: str
    sort: SortDirection = Field(default=SortDirection.ASCENDING)

    def __hash__(self) -> int:
        return hash((self.name, self.sort))

    def __eq__(self, other: Any) -> Condition:  # type: ignore[override]
        return Equals(prop=self, value=other)

    def __ne__(self, other: Any) -> Condition:  # type: ignore[override]
        return EqualsNot(prop=self, value=other)

    def __gt__(self, value: Any) -> Condition:
        return GreaterThan(prop=self, value=value)

    def __lt__(self, value: Any) -> Condition:
        return LessThan(prop=self, value=value)

    def __ge__(self, value: Any) -> Condition:
        return GreaterThanOrEqualTo(prop=self, value=value)

    def __le__(self, value: Any) -> Condition:
        return LessThanOrEqualTo(prop=self, value=value)

    def __repr__(self) -> str:
        return f"prop('{self.name}')"

    def __str__(self) -> str:
        return repr(self)

    def contains(self, value: str | User | Page | Option) -> Condition:
        return Contains(prop=self, value=value)

    def does_not_contain(self, value: str | User | Page | Option) -> Condition:
        return ContainsNot(prop=self, value=value)

    def starts_with(self, value: str) -> Condition:
        return StartsWith(prop=self, value=value)

    def ends_with(self, value: str) -> Condition:
        return EndsWith(prop=self, value=value)

    def is_empty(self) -> Condition:
        return IsEmpty(prop=self, value=True)

    def is_not_empty(self) -> Condition:
        return IsNotEmpty(prop=self, value=True)

    def this_week(self) -> Condition:
        return ThisWeek(prop=self, value=obj_query.DateCondition.EmptyObject())

    def past_week(self) -> Condition:
        return PastWeek(prop=self, value=obj_query.DateCondition.EmptyObject())

    def past_month(self) -> Condition:
        return PastMonth(prop=self, value=obj_query.DateCondition.EmptyObject())

    def past_year(self) -> Condition:
        return PastYear(prop=self, value=obj_query.DateCondition.EmptyObject())

    def next_week(self) -> Condition:
        return NextWeek(prop=self, value=obj_query.DateCondition.EmptyObject())

    def next_month(self) -> Condition:
        return NextMonth(prop=self, value=obj_query.DateCondition.EmptyObject())

    def next_year(self) -> Condition:
        return NextYear(prop=self, value=obj_query.DateCondition.EmptyObject())

    def asc(self) -> Self:
        self.sort = SortDirection.ASCENDING
        return self

    def desc(self) -> Self:
        self.sort = SortDirection.DESCENDING
        return self

    @property
    def any(self) -> RollupArrayProperty:
        return RollupArrayProperty(name=self.name, sort=self.sort, quantifier=ArrayQuantifier.ANY)

    @property
    def none(self) -> RollupArrayProperty:
        return RollupArrayProperty(name=self.name, sort=self.sort, quantifier=ArrayQuantifier.NONE)

    @property
    def every(self) -> RollupArrayProperty:
        return RollupArrayProperty(name=self.name, sort=self.sort, quantifier=ArrayQuantifier.EVERY)


class RollupArrayProperty(PageProperty):
    """Represents a rollup array property of a page."""

    quantifier: ArrayQuantifier

    def __repr__(self) -> str:
        return f"prop('{self.name}').{self.quantifier.value}"


def prop(prop_name: str, /) -> PageProperty:
    """Create a property object."""
    return PageProperty(name=prop_name)


class Condition(BaseModel, ABC):
    """Base class for filter query conditions."""

    is_method: bool = False  # to distinguish between method and operator conditions for parenthesizing

    def __and__(self, other: Condition) -> Condition:
        # This logic reduces the number of parentheses by combining ANDs (associative rule)
        match other:
            case And(terms=terms) if isinstance(self, And):
                return And(terms=[*self.terms, *terms])
            case And(terms=terms):
                return And(terms=[self, *terms])
            case _ if isinstance(self, And):
                return And(terms=[*self.terms, other])
            case _:
                return And(terms=[self, other])

    def __iand__(self, other: Condition) -> Condition:
        return self & other

    def __or__(self, other: Condition) -> Condition:
        # This logic reduces the number of parentheses by combining ORs (associative rule)
        match other:
            case Or(terms=terms) if isinstance(self, Or):
                return Or(terms=[*self.terms, *terms])
            case Or(terms=terms):
                return Or(terms=[self, *terms])
            case _ if isinstance(self, Or):
                return Or(terms=[*self.terms, other])
            case _:
                return Or(terms=[self, other])

    def __ior__(self, other: Condition) -> Condition:
        return self | other

    @abstractmethod
    def __repr__(self) -> str: ...

    def __str__(self) -> str:
        return repr(self)

    @abstractmethod
    def create_obj_ref(self, db: Database) -> obj_query.QueryFilter: ...


class PropertyCondition(Condition, ABC):
    prop: PageProperty
    value: Any
    _probe_page: Page | None = None

    @abstractmethod
    def _create_obj_ref_kwargs(self, db: Database, prop_type: Property) -> dict[str, obj_query.Condition]:
        """Create the keyword arguments for the obj_query.PropertyFilter constructor.

        We need this as a rollup array condition works on top of a property condition.
        Thus we handle here everything except of the rollup array condition.
        """

    def _get_prop_type(self, db: Database) -> Property:
        return db.schema[self.prop.name]

    def _get_probe_page(self, db: Database) -> Page:
        """Return a page from the database to probe its properties as needed for certain conditions."""
        if self._probe_page is None:
            session = get_active_session()
            try:
                page_obj = next(session.api.databases.query(db.obj_ref).execute(page_size=1))
            except StopIteration as e:
                msg = f'The database {db} is empty.'
                raise EmptyDBError(msg) from e
            page_obj_id = raise_unset(page_obj.id)
            self._probe_page = cast(Page, session.cache.setdefault(page_obj_id, Page.wrap_obj_ref(page_obj)))

        return self._probe_page

    def _get_formula_type(self, db: Database) -> FormulaType:
        """Return the type of a formula property."""
        page = self._get_probe_page(db)
        prop: props.PropertyValue = props.PropertyValue.wrap_obj_ref(page.props._obj_prop_vals[self.prop.name])

        if not isinstance(prop, props.Formula):
            msg = f'The property {self.prop.name} is not a formula property.'
            raise FilterQueryError(msg)

        if not (prop_type := prop.value_type):
            msg = f'The property {self.prop.name} does not have a formula type set.'
            raise FilterQueryError(msg)

        return prop_type

    def _get_rollup_type(self, db: Database) -> RollupType:
        """Return the type of a rollup property."""
        page = self._get_probe_page(db)
        prop: props.PropertyValue = props.PropertyValue.wrap_obj_ref(page.props._obj_prop_vals[self.prop.name])

        if isinstance(prop, props.Rollup) and (prop_type := prop.value_type) is not None:
            return prop_type
        else:
            msg = f'The property {self.prop.name} is not a rollup property or is missing a type.'
            raise FilterQueryError(msg)

    @property
    def _value_str(self) -> str:
        match self.value:
            case str() | dt.datetime() | Option():
                return f"'{self.value}'"
            case _:
                return f'{self.value}'


class IsEmpty(PropertyCondition):
    _condition_kw = 'is_empty'
    is_method: bool = True

    def _create_obj_ref_kwargs(self, db: Database, prop_type: Property) -> dict[str, obj_query.Condition]:
        kwargs: dict[str, obj_query.Condition] = {}

        match prop_type:
            case schema.Text() | schema.Title() | schema.Phone() | schema.Email() | schema.URL():
                kwargs['rich_text'] = obj_query.TextCondition(**{self._condition_kw: self.value})
            case schema.Number():
                kwargs['number'] = obj_query.NumberCondition(**{self._condition_kw: self.value})
            case schema.Select():
                kwargs['select'] = obj_query.SelectCondition(**{self._condition_kw: str(self.value)})
            case schema.MultiSelect():
                kwargs['multi_select'] = obj_query.MultiSelectCondition(**{self._condition_kw: str(self.value)})
            case schema.Date():
                kwargs['date'] = obj_query.DateCondition(**{self._condition_kw: self.value})
            case schema.Person():
                kwargs['people'] = obj_query.PeopleCondition(**{self._condition_kw: self.value})
            case schema.Files():
                kwargs['files'] = obj_query.FilesCondition(**{self._condition_kw: self.value})
            case schema.Relation():
                kwargs['relation'] = obj_query.RelationCondition(**{self._condition_kw: self.value})
            case schema.Formula():
                condition: obj_query.Condition

                match formula_type := self._get_formula_type(db):
                    case FormulaType.STRING:
                        condition = obj_query.TextCondition(**{self._condition_kw: self.value})
                    case FormulaType.NUMBER:
                        condition = obj_query.NumberCondition(**{self._condition_kw: self.value})
                    case FormulaType.DATE:
                        condition = obj_query.DateCondition(**{self._condition_kw: self.value})
                    case _:
                        msg = f'Invalid formula type `{formula_type.value}` for condition {self}.'
                        raise FilterQueryError(msg)

                kwargs['formula'] = obj_query.FormulaCondition(**{formula_type.formula_kwarg: condition})
            case _:
                msg = f'Invalid property type `{prop_type}` for condition {self}.'
                raise FilterQueryError(msg)

        return kwargs

    def create_obj_ref(self, db: Database) -> obj_query.QueryFilter:
        prop_type = self._get_prop_type(db)

        match prop_type:
            case schema.CreatedTime():
                date_condition = obj_query.DateCondition(**{self._condition_kw: self.value})
                return obj_query.CreatedTimeFilter(created_time=date_condition)
            case schema.LastEditedTime():
                date_condition = obj_query.DateCondition(**{self._condition_kw: self.value})
                return obj_query.LastEditedTimeFilter(last_edited_time=date_condition)
            case schema.Rollup():
                condition: obj_query.Condition

                match rollup_type := self._get_rollup_type(db):
                    case RollupType.ARRAY:
                        if not isinstance(self.prop, RollupArrayProperty):
                            msg = (
                                f'The property {self.prop.name} must be a rollup array property, '
                                'use one of the properties `any`, `every` and `none`.'
                            )
                            raise FilterQueryError(msg)

                        kwargs = self._create_obj_ref_kwargs(db, prop_type.rollup_prop)
                        condition = obj_query.RollupArrayCondition(**kwargs)
                        rollup_kwarg = self.prop.quantifier.value
                    case RollupType.NUMBER:
                        condition = obj_query.NumberCondition(**{self._condition_kw: self.value})
                        rollup_kwarg = rollup_type.value
                    case RollupType.DATE:
                        condition = obj_query.DateCondition(**{self._condition_kw: self.value})
                        rollup_kwarg = rollup_type.value
                    case _:
                        msg = f'Invalid rollup type `{rollup_type}` for condition {self}.'
                        raise FilterQueryError(msg)

                kwargs = {'rollup': obj_query.RollupCondition(**{rollup_kwarg: condition})}
            case _:
                kwargs = self._create_obj_ref_kwargs(db, prop_type)

        return obj_query.PropertyFilter(property=self.prop.name, **kwargs)

    def __repr__(self) -> str:
        return f'{self.prop}.{self._condition_kw}()'


class IsNotEmpty(IsEmpty):
    _condition_kw = 'is_not_empty'
    is_method: bool = True


class Equals(PropertyCondition):
    _condition_kw = 'equals'

    def _create_obj_ref_kwargs(self, db: Database, prop_type: Property) -> dict[str, obj_query.Condition]:
        kwargs: dict[str, obj_query.Condition] = {}

        match prop_type:
            case schema.Text() | schema.Title() | schema.Phone() | schema.Email() | schema.URL():
                kwargs['rich_text'] = obj_query.TextCondition(**{self._condition_kw: self.value})
            case schema.ID():
                kwargs['unique_id'] = obj_query.IdCondition(**{self._condition_kw: self.value})
            case schema.Number():
                kwargs['number'] = obj_query.NumberCondition(**{self._condition_kw: self.value})
            case schema.Checkbox():
                kwargs['checkbox'] = obj_query.CheckboxCondition(**{self._condition_kw: bool(self.value)})
            case schema.Select():
                kwargs['select'] = obj_query.SelectCondition(**{self._condition_kw: str(self.value)})
            case schema.Date() if self._condition_kw == 'equals':  # date has no `does_not_equal` condition
                kwargs['date'] = obj_query.DateCondition(**{self._condition_kw: self.value})
            case schema.Formula():
                condition: obj_query.Condition

                match formula_type := self._get_formula_type(db):
                    case FormulaType.STRING:
                        condition = obj_query.TextCondition(**{self._condition_kw: self.value})
                    case FormulaType.NUMBER:
                        condition = obj_query.NumberCondition(**{self._condition_kw: self.value})
                    case FormulaType.DATE:
                        condition = obj_query.DateCondition(**{self._condition_kw: self.value})
                    case FormulaType.BOOLEAN:
                        condition = obj_query.CheckboxCondition(**{self._condition_kw: bool(self.value)})

                kwargs['formula'] = obj_query.FormulaCondition(**{formula_type.formula_kwarg: condition})

            case _:
                msg = f'Invalid property type `{prop_type}` for condition {self}.'
                raise FilterQueryError(msg)

        return kwargs

    def create_obj_ref(self, db: Database) -> obj_query.QueryFilter:
        prop_type = self._get_prop_type(db)

        match prop_type:
            case schema.CreatedTime():
                date_condition = obj_query.DateCondition(**{self._condition_kw: self.value})
                return obj_query.CreatedTimeFilter(created_time=date_condition)
            case schema.LastEditedTime():
                date_condition = obj_query.DateCondition(**{self._condition_kw: self.value})
                return obj_query.LastEditedTimeFilter(last_edited_time=date_condition)
            case schema.Rollup():
                condition: obj_query.Condition

                match rollup_type := self._get_rollup_type(db):
                    case RollupType.ARRAY:
                        if not isinstance(self.prop, RollupArrayProperty):
                            msg = (
                                f'The property {self.prop.name} must be a rollup array property, '
                                'use one of the properties `any`, `every` and `none`.'
                            )
                            raise FilterQueryError(msg)

                        kwargs = self._create_obj_ref_kwargs(db, prop_type.rollup_prop)
                        condition = obj_query.RollupArrayCondition(**kwargs)
                        rollup_kwarg = self.prop.quantifier.value
                    case RollupType.NUMBER:
                        condition = obj_query.NumberCondition(**{self._condition_kw: self.value})
                        rollup_kwarg = rollup_type.value
                    case RollupType.DATE:
                        condition = obj_query.DateCondition(**{self._condition_kw: self.value})
                        rollup_kwarg = rollup_type.value
                    case _:
                        msg = f'Invalid rollup type `{rollup_type}` for condition {self}.'
                        raise FilterQueryError(msg)

                kwargs = {'rollup': obj_query.RollupCondition(**{rollup_kwarg: condition})}
            case _:
                kwargs = self._create_obj_ref_kwargs(db, prop_type)

        return obj_query.PropertyFilter(property=self.prop.name, **kwargs)

    def __repr__(self) -> str:
        return f'{self.prop} == {self._value_str}'


class EqualsNot(Equals):
    _condition_kw = 'does_not_equal'

    def __repr__(self) -> str:
        return f'{self.prop} != {self._value_str}'


class InEquality(PropertyCondition, ABC):
    _num_condition_kw: str
    _date_condition_kw: str

    def _create_obj_ref_kwargs(self, db: Database, prop_type: Property) -> dict[str, obj_query.Condition]:
        kwargs: dict[str, obj_query.Condition] = {}

        match prop_type:
            case schema.Number():
                kwargs['number'] = obj_query.NumberCondition(**{self._num_condition_kw: self.value})
            case schema.ID():
                kwargs['unique_id'] = obj_query.IdCondition(**{self._num_condition_kw: self.value})
            case schema.Date():
                kwargs['date'] = obj_query.DateCondition(**{self._date_condition_kw: self.value})
            case schema.Formula():
                condition: obj_query.Condition

                match formula_type := self._get_formula_type(db):
                    case FormulaType.NUMBER:
                        condition = obj_query.NumberCondition(**{self._num_condition_kw: self.value})
                    case FormulaType.DATE:
                        condition = obj_query.DateCondition(**{self._date_condition_kw: self.value})
                    case _:
                        msg = f'Invalid formula type `{formula_type.value}` for condition {self}.'
                        raise FilterQueryError(msg)

                kwargs['formula'] = obj_query.FormulaCondition(**{formula_type.formula_kwarg: condition})
            case _:
                msg = f'Invalid property type `{prop_type}` for condition {self}.'
                raise FilterQueryError(msg)

        return kwargs

    def create_obj_ref(self, db: Database) -> obj_query.QueryFilter:
        prop_type = self._get_prop_type(db)

        match prop_type:
            case schema.CreatedTime():
                date_condition = obj_query.DateCondition(**{self._date_condition_kw: self.value})
                return obj_query.CreatedTimeFilter(created_time=date_condition)
            case schema.LastEditedTime():
                date_condition = obj_query.DateCondition(**{self._date_condition_kw: self.value})
                return obj_query.LastEditedTimeFilter(last_edited_time=date_condition)
            case schema.Rollup():
                condition: obj_query.Condition

                match rollup_type := self._get_rollup_type(db):
                    case RollupType.ARRAY:
                        if not isinstance(self.prop, RollupArrayProperty):
                            msg = (
                                f'The property {self.prop.name} must be a rollup array property, '
                                'use one of the properties `any`, `every` and `none`.'
                            )
                            raise FilterQueryError(msg)

                        kwargs = self._create_obj_ref_kwargs(db, prop_type.rollup_prop)
                        condition = obj_query.RollupArrayCondition(**kwargs)
                        rollup_kwarg = self.prop.quantifier.value
                    case RollupType.NUMBER:
                        condition = obj_query.NumberCondition(**{self._num_condition_kw: self.value})
                        rollup_kwarg = rollup_type.value
                    case RollupType.DATE:
                        condition = obj_query.DateCondition(**{self._date_condition_kw: self.value})
                        rollup_kwarg = rollup_type.value
                    case _:
                        msg = f'Invalid rollup type `{rollup_type}` for condition {self}.'
                        raise FilterQueryError(msg)

                kwargs = {'rollup': obj_query.RollupCondition(**{rollup_kwarg: condition})}
            case _:
                kwargs = self._create_obj_ref_kwargs(db, prop_type)

        return obj_query.PropertyFilter(property=self.prop.name, **kwargs)

    @abstractmethod
    def __repr__(self) -> str: ...


class GreaterThan(InEquality):
    _num_condition_kw = 'greater_than'
    _date_condition_kw = 'after'

    def __repr__(self) -> str:
        return f'{self.prop} > {self._value_str}'


class LessThan(InEquality):
    _num_condition_kw = 'less_than'
    _date_condition_kw = 'before'

    def __repr__(self) -> str:
        return f'{self.prop} < {self._value_str}'


class GreaterThanOrEqualTo(InEquality):
    _num_condition_kw = 'greater_than_or_equal_to'
    _date_condition_kw = 'on_or_after'

    def __repr__(self) -> str:
        return f'{self.prop} >= {self._value_str}'


class LessThanOrEqualTo(InEquality):
    _num_condition_kw = 'less_than_or_equal_to'
    _date_condition_kw = 'on_or_before'

    def __repr__(self) -> str:
        return f'{self.prop} <= {self._value_str}'


class Contains(PropertyCondition):
    _condition_kw = 'contains'
    is_method: bool = True

    def _create_obj_ref_kwargs(self, db: Database, prop_type: Property) -> dict[str, obj_query.Condition]:
        kwargs: dict[str, obj_query.Condition] = {}

        match prop_type:
            case schema.Text() | schema.Title() | schema.Phone() | schema.Email() | schema.URL():
                kwargs['rich_text'] = obj_query.TextCondition(**{self._condition_kw: self.value})
            case schema.MultiSelect():
                kwargs['multi_select'] = obj_query.MultiSelectCondition(**{self._condition_kw: str(self.value)})
            case schema.Person() if isinstance(self.value, User):
                kwargs['people'] = obj_query.PeopleCondition(**{self._condition_kw: self.value.id})
            case schema.Relation() if isinstance(self.value, Page):
                kwargs['relation'] = obj_query.RelationCondition(**{self._condition_kw: self.value.id})
            case schema.Formula():
                condition: obj_query.Condition

                match formula_type := self._get_formula_type(db):
                    case FormulaType.STRING:
                        condition = obj_query.TextCondition(**{self._condition_kw: self.value})
                    case _:
                        msg = f'Invalid formula type `{formula_type.value}` for condition {self}.'
                        raise FilterQueryError(msg)

                kwargs['formula'] = obj_query.FormulaCondition(**{formula_type.formula_kwarg: condition})
            case _:
                msg = f'Invalid property type `{prop_type}` for condition {self}.'
                raise FilterQueryError(msg)

        return kwargs

    def create_obj_ref(self, db: Database) -> obj_query.QueryFilter:
        prop_type = self._get_prop_type(db)

        match prop_type:
            case schema.Rollup():
                match rollup_type := self._get_rollup_type(db):
                    case RollupType.ARRAY:
                        if not isinstance(self.prop, RollupArrayProperty):
                            msg = (
                                f'The property {self.prop.name} must be a rollup array property, '
                                'use one of the properties `any`, `every` and `none`.'
                            )
                            raise FilterQueryError(msg)

                        kwargs = self._create_obj_ref_kwargs(db, prop_type.rollup_prop)
                        condition = obj_query.RollupArrayCondition(**kwargs)
                    case _:
                        msg = f'Invalid rollup type `{rollup_type}` for condition {self}.'
                        raise FilterQueryError(msg)

                kwargs = {'rollup': obj_query.RollupCondition(**{self.prop.quantifier.value: condition})}
            case _:
                kwargs = self._create_obj_ref_kwargs(db, prop_type)

        return obj_query.PropertyFilter(property=self.prop.name, **kwargs)

    def __repr__(self) -> str:
        return f'{self.prop}.{self._condition_kw}({self._value_str})'


class ContainsNot(Contains):
    _condition_kw = 'does_not_contain'
    is_method: bool = True


class StartsWith(PropertyCondition):
    _condition_kw = 'starts_with'
    is_method: bool = True

    def _create_obj_ref_kwargs(self, db: Database, prop_type: Property) -> dict[str, obj_query.Condition]:
        kwargs: dict[str, obj_query.Condition] = {}

        match prop_type:
            case schema.Text() | schema.Title() | schema.Phone() | schema.Email() | schema.URL():
                kwargs['rich_text'] = obj_query.TextCondition(**{self._condition_kw: self.value})
            case schema.Formula():
                match formula_type := self._get_formula_type(db):
                    case FormulaType.STRING:
                        condition = obj_query.TextCondition(**{self._condition_kw: self.value})
                    case _:
                        msg = f'Invalid formula type `{formula_type.value}` for condition {self}.'
                        raise FilterQueryError(msg)

                kwargs['formula'] = obj_query.FormulaCondition(**{formula_type.formula_kwarg: condition})
            case _:
                msg = f'Invalid property type `{prop_type}` for condition {self}.'
                raise FilterQueryError(msg)

        return kwargs

    def create_obj_ref(self, db: Database) -> obj_query.QueryFilter:
        prop_type = self._get_prop_type(db)

        match prop_type:
            case schema.Rollup():
                match rollup_type := self._get_rollup_type(db):
                    case RollupType.ARRAY:
                        if not isinstance(self.prop, RollupArrayProperty):
                            msg = (
                                f'The property {self.prop.name} must be a rollup array property, '
                                'use one of the properties `any`, `every` and `none`.'
                            )
                            raise FilterQueryError(msg)

                        kwargs = self._create_obj_ref_kwargs(db, prop_type.rollup_prop)
                        condition = obj_query.RollupArrayCondition(**kwargs)
                    case _:
                        msg = f'Invalid rollup type `{rollup_type}` for condition {self}.'
                        raise FilterQueryError(msg)

                kwargs = {'rollup': obj_query.RollupCondition(**{self.prop.quantifier.value: condition})}
            case _:
                kwargs = self._create_obj_ref_kwargs(db, prop_type)

        return obj_query.PropertyFilter(property=self.prop.name, **kwargs)

    def __repr__(self) -> str:
        return f'{self.prop}.{self._condition_kw}({self._value_str})'


class EndsWith(StartsWith):
    _condition_kw = 'ends_with'
    is_method: bool = True


class DateCondition(PropertyCondition, ABC):
    _condition_kw: str
    is_method: bool = True

    def _create_obj_ref_kwargs(self, db: Database, prop_type: Property) -> dict[str, obj_query.Condition]:
        kwargs: dict[str, obj_query.Condition] = {}

        match prop_type:
            case schema.Date():
                kwargs['date'] = obj_query.DateCondition(**{self._condition_kw: self.value})
            case schema.Formula():
                match formula_type := self._get_formula_type(db):
                    case FormulaType.DATE:
                        condition = obj_query.DateCondition(**{self._condition_kw: self.value})
                    case _:
                        msg = f'Invalid formula type `{formula_type.value}` for condition {self}.'
                        raise FilterQueryError(msg)

                kwargs['formula'] = obj_query.FormulaCondition(**{formula_type.formula_kwarg: condition})
            case _:
                msg = f'Invalid property type `{prop_type}` for condition {self}.'
                raise FilterQueryError(msg)

        return kwargs

    def create_obj_ref(self, db: Database) -> obj_query.QueryFilter:
        prop_type = self._get_prop_type(db)

        match prop_type:
            case schema.CreatedTime():
                date_condition = obj_query.DateCondition(**{self._condition_kw: self.value})
                return obj_query.CreatedTimeFilter(created_time=date_condition)
            case schema.LastEditedTime():
                date_condition = obj_query.DateCondition(**{self._condition_kw: self.value})
                return obj_query.LastEditedTimeFilter(last_edited_time=date_condition)
            case schema.Rollup():
                condition: obj_query.Condition

                match rollup_type := self._get_rollup_type(db):
                    case RollupType.ARRAY:
                        if not isinstance(self.prop, RollupArrayProperty):
                            msg = (
                                f'The property {self.prop.name} must be a rollup array property, '
                                'use one of the properties `any`, `every` and `none`.'
                            )
                            raise FilterQueryError(msg)

                        kwargs = self._create_obj_ref_kwargs(db, prop_type.rollup_prop)
                        condition = obj_query.RollupArrayCondition(**kwargs)
                        rollup_kwarg = self.prop.quantifier.value
                    case RollupType.DATE:
                        condition = obj_query.DateCondition(**{self._condition_kw: self.value})
                        rollup_kwarg = rollup_type.value
                    case _:
                        msg = f'Invalid rollup type `{rollup_type}` for condition {self}.'
                        raise FilterQueryError(msg)

                kwargs = {'rollup': obj_query.RollupCondition(**{rollup_kwarg: condition})}
            case _:
                kwargs = self._create_obj_ref_kwargs(db, prop_type)

        return obj_query.PropertyFilter(property=self.prop.name, **kwargs)

    def __repr__(self) -> str:
        return f'{self.prop}.{self._condition_kw}()'


class PastWeek(DateCondition):
    _condition_kw = 'past_week'


class PastMonth(DateCondition):
    _condition_kw = 'past_month'


class PastYear(DateCondition):
    _condition_kw = 'past_year'


class ThisWeek(DateCondition):
    _condition_kw = 'this_week'


class NextWeek(DateCondition):
    _condition_kw = 'next_week'


class NextMonth(DateCondition):
    _condition_kw = 'next_month'


class NextYear(DateCondition):
    _condition_kw = 'next_year'


class And(Condition):
    terms: list[Condition]

    def create_obj_ref(self, db: Database) -> obj_query.QueryFilter:
        return obj_query.CompoundFilter(and_=[term.create_obj_ref(db) for term in self.terms])

    def __repr__(self) -> str:
        terms = [f'({term})' if not term.is_method else str(term) for term in self.terms]
        return ' & '.join(str(term) for term in terms)


class Or(Condition):
    terms: list[Condition]

    def create_obj_ref(self, db: Database) -> obj_query.QueryFilter:
        return obj_query.CompoundFilter(or_=[term.create_obj_ref(db) for term in self.terms])

    def __repr__(self) -> str:
        terms = [f'({term})' if not term.is_method else str(term) for term in self.terms]
        return ' | '.join(str(term) for term in terms)


class Query:
    """A query object to filter and sort pages in a database."""

    database: Database
    _filter: Condition | None = None
    _sorts: list[PageProperty]

    def __init__(self, database: Database):
        self.database = database
        self._sorts = []

    def _filter_obj_ref(self) -> obj_query.QueryFilter | None:
        if self._filter is None:
            return None
        else:
            return self._filter.create_obj_ref(self.database)

    def _sorts_obj_ref(self) -> list[obj_query.DBSort]:
        return [obj_query.DBSort(property=prop.name, direction=prop.sort) for prop in self._sorts]

    def execute(self) -> View:
        """Execute the query and return the resulting pages as a view."""
        sorts = ', '.join(f'{prop}.{prop.sort}()' for prop in self._sorts) if self._sorts else ''
        _logger.info(f'Querying database `{self.database}` with filter `{self._filter}` and sorts `{sorts}`.')
        session = get_active_session()
        query_obj = session.api.databases.query(self.database.obj_ref)

        try:
            if filter_obj := self._filter_obj_ref():
                query_obj = query_obj.filter(filter_obj)
        except EmptyDBError:
            return View(database=self.database, pages=[], query=self)

        if sort_objs := self._sorts_obj_ref():
            query_obj = query_obj.sort(sort_objs)

        pages = [
            cast(Page, session.cache.setdefault(raise_unset(page.id), Page.wrap_obj_ref(page)))
            for page in query_obj.execute()
        ]
        return View(database=self.database, pages=pages, query=self)

    def filter(self, expr: Condition) -> Query:
        """Filter the query by the given properties.

        !!! note
            The filter is applied as an AND operation with the existing filter.
        """
        if self._filter is None:
            self._filter = expr
        else:
            self._filter &= expr

        return self

    def sort(self, *props: PageProperty | str) -> Query:
        """Sort the query by the given properties and directions.

        !!! note
            The order of the properties is important. The first property is the primary sort,
            the second is the secondary sort, and so on. Calling this method multiple times
            will overwrite the previous sorts.
        """
        self._sorts = [prop if isinstance(prop, PageProperty) else PageProperty(name=prop) for prop in props]
        return self

    def __repr__(self) -> str:
        sorts = ', '.join(f'{prop}.{prop.sort}()' for prop in self._sorts) if self._sorts else ''
        return f"Query(database='{self.database}', sort=({sorts}), filter={self._filter})"

    def __str__(self) -> str:
        return repr(self)



"""Functions and classes for working with users in Notion."""

from __future__ import annotations

from typing import TYPE_CHECKING

from typing_extensions import TypeVar

from ultimate_notion.core import Wrapper, get_repr
from ultimate_notion.obj_api import objects as objs

if TYPE_CHECKING:
    from uuid import UUID


class WorkSpaceInfo(objs.WorkSpaceLimits):
    """Workspace information for a bot user."""

    name: str


U_co = TypeVar('U_co', bound=objs.User, default=objs.User, covariant=True)


class User(Wrapper[U_co], wraps=objs.User):
    """User object for persons, bots and unknown users.

    Unknown users are users, which no longer participate in the workspace
    or were revoked access. They are represented by their ID and have
    the name `Unknown User`.
    """

    def __str__(self) -> str:
        return self.name or f'Unnamed user {self.id}>'

    def __repr__(self) -> str:
        return get_repr(self)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, User):
            return NotImplemented
        return self.id == other.id

    def __hash__(self) -> int:
        return hash(self.id)

    @property
    def id(self) -> UUID:
        """Return the ID of this user."""
        return self.obj_ref.id

    @property
    def name(self) -> str | None:
        """Return the name of this user."""
        return self.obj_ref.name

    @property
    def avatar_url(self) -> str | None:
        """Return the avatar URL of this user."""
        return self.obj_ref.avatar_url

    @property
    def is_person(self) -> bool:
        """Return True if this user is a person."""
        return False

    @property
    def is_bot(self) -> bool:
        """Return True if this user is a bot."""
        return False

    @property
    def is_unknown(self) -> bool:
        """Return True if this user is an unknown user."""
        return False


class Person(User[objs.Person], wraps=objs.Person):
    """A user that represents a person."""

    @property
    def is_person(self) -> bool:
        return True

    @property
    def email(self) -> str | None:
        """Return the e-mail address of this user, if available."""
        if isinstance(self.obj_ref, objs.Person):
            return self.obj_ref.person.email
        else:  # it's a bot or unknown without an e-mail
            return None


class Bot(User[objs.Bot], wraps=objs.Bot):
    """A user that represents a bot."""

    @property
    def is_bot(self) -> bool:
        return True

    @property
    def workspace_info(self) -> WorkSpaceInfo:
        """Return the workspace info of this bot, if available."""
        kwargs = self.obj_ref.bot.workspace_limits.model_dump()
        return WorkSpaceInfo(name=self.obj_ref.bot.workspace_name, **kwargs)


class UnknownUser(User[objs.UnknownUser], wraps=objs.UnknownUser):
    """A user that is unknown, i.e. no longer part of the workspace."""

    def __str__(self) -> str:
        return self.name or f'Unknown user {self.id}>'

    @property
    def is_unknown(self) -> bool:
        return True



"""Wrapper for various Notion API objects like parents, mentions, emojis & users.

Similar to other records, these object provide access to the primitive data structure
used in the Notion API.

For validation the Pydantic model fields specify if a field is optional or not.
Some fields are always set, e.g. `id`, when retrieving an object but must not be set
when sending the object to the Notion API in order to create the object.
To model this behavior, the default sentinel value `Unset` is used for those objects, e.g.
```
class SelectOption(GenericObject)
    id: str | UnsetType = Unset
```
Be aware that this is important when updating to differentiate between the actual set
values from default/unset values.
"""

from __future__ import annotations

import datetime as dt
from abc import ABC, abstractmethod
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Literal
from uuid import UUID

import pendulum as pnd
from pydantic import Field, SerializeAsAny
from typing_extensions import TypeVar

from ultimate_notion.obj_api.core import (
    GenericObject,
    NotionEntity,
    NotionObject,
    TypedObject,
    Unset,
    UnsetType,
    _normalize_color,
    extract_id,
    raise_unset,
)
from ultimate_notion.obj_api.enums import BGColor, Color, FileUploadStatus
from ultimate_notion.utils import DateTimeOrRange, parse_dt_str

if TYPE_CHECKING:
    from ultimate_notion.obj_api.blocks import Block, Database, Page


class SelectOption(GenericObject):
    """Options for select & multi-select objects.

    Specifying no color will result in the default color being used, i.e. `Color.DEFAULT`, which is a light grey.
    Note that colors can't be changed after they are set.
    """

    name: str
    id: str | UnsetType = Unset  # According to docs: "These are sometimes, but not always, UUIDs."
    color: Color | UnsetType = Unset  # Leave this as Unset when overwriting an option as colors can't be changed
    description: list[RichTextBaseObject] | None = None  # ToDo: Undocumented in the Notion API

    @classmethod
    def build(cls, name: str, color: Color = Color.DEFAULT) -> SelectOption:
        """Create a `SelectOption` object from the given name and color."""
        return cls.model_construct(name=name, color=color)

    def __eq__(self, other: object) -> bool:
        """Compare SelectOption objects by all attributes except id."""
        if not isinstance(other, SelectOption):
            return False

        my_color, other_color = _normalize_color(self.color), _normalize_color(other.color)
        return self.name == other.name and my_color == other_color and self.description == other.description

    def __hash__(self) -> int:
        """Return a hash of the SelectOption based on name, color, and description."""
        # Convert description list to tuple for hashing if it exists
        desc_tuple = tuple(self.description) if self.description else None
        return hash((self.name, self.color, desc_tuple))


class SelectGroup(GenericObject):
    """Group of options for status objects."""

    name: str
    id: str | UnsetType = Unset  # According to docs: "These are sometimes, but not always, UUIDs."
    color: Color | UnsetType = Unset
    option_ids: list[str] = Field(default_factory=list)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SelectGroup):
            return False
        # we don't compare option_ids since we might not always have them (Status property can't be created)
        my_color, other_color = _normalize_color(self.color), _normalize_color(other.color)
        return self.name == other.name and my_color == other_color

    def __hash__(self) -> int:
        return hash((self.name, self.color))


class MentionMixin(ABC):
    """Mixin for objects that can be mentioned in Notion.

    This mixin adds a `build_mention` property to the object, which can be used to
    reference the object in a mention.
    """

    @abstractmethod
    def build_mention(self, style: Annotations | None = None) -> MentionObject:
        """Return a mention object for this object."""


class DateRange(GenericObject, MentionMixin):
    """A Notion date range, with an optional end date."""

    start: dt.date | dt.datetime
    end: dt.date | dt.datetime | None = None
    time_zone: str | None = None

    @classmethod
    def build(cls, dt_spec: str | DateTimeOrRange) -> DateRange:
        """Compose a DateRange object from the given properties."""

        if isinstance(dt_spec, str):
            dt_spec = parse_dt_str(dt_spec)

        match dt_spec:
            case pnd.Interval():
                return cls.model_construct(
                    time_zone=dt_spec.start.timezone_name if isinstance(dt_spec.start, pnd.DateTime) else None,
                    start=dt_spec.start,
                    end=dt_spec.end,
                )

            case pnd.DateTime():
                return cls.model_construct(time_zone=dt_spec.timezone_name, start=dt_spec.naive(), end=None)

            case dt.datetime():
                # We don't trust the timezone of naive datetime and convert to UTC
                return cls.model_construct(time_zone='UTC', start=dt_spec.astimezone(dt.timezone.utc), end=None)

            case pnd.Date() | dt.date():
                return cls.model_construct(time_zone=None, start=dt_spec, end=None)

            case _:
                msg = f"Unsupported type for 'dt_spec': {type(dt_spec)}"
                raise TypeError(msg)

    def build_mention(self, style: Annotations | None = None) -> MentionObject:
        return MentionDate.build_mention_from(self, style=style)

    def to_pendulum(self) -> DateTimeOrRange:
        """Convert the DateRange to a pendulum object."""
        # self.time_zone is None for pure dates.
        start = str(self.start) if self.time_zone is None else f'{self.start} {self.time_zone}'
        if self.end is None:
            return parse_dt_str(start)
        else:
            end = str(self.end) if self.time_zone is None else f'{self.end} {self.time_zone}'
            start_dt, end_dt = parse_dt_str(start), parse_dt_str(end)

            if isinstance(start_dt, pnd.Date) and isinstance(end_dt, pnd.Date):
                return pnd.Interval(start=start_dt, end=end_dt)
            elif isinstance(start_dt, pnd.DateTime) and isinstance(end_dt, pnd.DateTime):
                return pnd.Interval(start=start_dt, end=end_dt)
            else:
                msg = f"Unsupported type for 'start' or 'end': {type(start)}, {type(end)}"
                raise TypeError(msg)

    def __str__(self) -> str:
        # ToDo: Implement the possibility to configure date format globally, maybe in the config?
        date_str = str(self.start) if self.end is None else f'{self.start} → {self.end}'
        return f'@{date_str}'  # we add the @ as Notion does it too


class LinkObject(GenericObject):
    """Reference a URL."""

    type: str = 'url'
    url: str


class ObjectRef(GenericObject):
    """A general-purpose object reference in the Notion API."""

    id: UUID

    @classmethod
    def build(cls, ref: ParentRef | GenericObject | UUID | str) -> ObjectRef:
        """Compose a reference to an object from the given reference.

        Strings may be either UUID's or URL's to Notion content.
        """

        match ref:
            case ObjectRef():
                return ref.model_copy(deep=True)
            case ParentRef() if isinstance(ref.value, UUID):
                # ParentRefs are typed objects with a nested UUID
                return cls.model_construct(id=ref.value)
            case GenericObject() if hasattr(ref, 'id'):
                # Re-compose the ObjectReference from the internal ID
                return cls.build(ref.id)
            case UUID():
                return cls.model_construct(id=ref)
            case str() if (id_str := extract_id(ref)) is not None:
                return cls.model_construct(id=UUID(id_str))
            case _:
                msg = f'Cannot interpret {ref} of type {type(ref)} as a reference to an object.'
                raise ValueError(msg)


def get_uuid(obj: str | UUID | ParentRef | NotionObject | BlockRef) -> UUID:
    """Retrieves a UUID from an object reference.

    Only meant for internal use.
    """
    return ObjectRef.build(obj).id


T = TypeVar('T', default=Any)


class ParentRef(TypedObject[T], ABC, polymorphic_base=True):
    """Reference another block as a parent.

    Notion API: [Parent Object](https://developers.notion.com/reference/parent-object)

    Note: This class is simply a placeholder for the typed concrete *Ref classes.
          Callers should always instantiate the intended concrete versions.
    """


class DatabaseRef(ParentRef[UUID], type='database_id'):
    """Reference a database."""

    database_id: UUID

    @classmethod
    def build(cls, db_ref: Database | str | UUID) -> DatabaseRef:
        """Compose a DatabaseRef from the given reference object."""
        ref = ObjectRef.build(db_ref)
        return DatabaseRef.model_construct(database_id=ref.id)


class PageRef(ParentRef[UUID], type='page_id'):
    """Reference a page."""

    page_id: UUID

    @classmethod
    def build(cls, page_ref: Page | str | UUID) -> PageRef:
        """Compose a PageRef from the given reference object."""
        ref = ObjectRef.build(page_ref)
        return PageRef.model_construct(page_id=ref.id)


class BlockRef(ParentRef[UUID], type='block_id'):
    """Reference a block."""

    block_id: UUID

    @classmethod
    def build(cls, block_ref: Block | str | UUID) -> BlockRef:
        """Compose a BlockRef from the given reference object."""
        ref = ObjectRef.build(block_ref)
        return BlockRef.model_construct(block_id=ref.id)


class WorkspaceRef(ParentRef[bool], type='workspace'):
    """Reference the workspace."""

    workspace: bool = True


class CommentRef(ParentRef[UUID], type='comment_id'):
    """Reference a comment."""

    comment_id: UUID

    @classmethod
    def build(cls, comment_ref: Comment | str | UUID) -> CommentRef:
        """Compose a CommentRef from the given reference object."""
        ref = ObjectRef.build(comment_ref)
        return CommentRef.model_construct(comment_id=ref.id)


class UserRef(NotionObject, object='user'):
    """Reference to a user, e.g. in `created_by`, `last_edited_by`, mentioning, etc."""

    id: UUID

    @classmethod
    def build(cls, user_ref: User | str | UUID) -> UserRef:
        """Compose a PageRef from the given reference object."""
        ref = ObjectRef.build(user_ref)
        return UserRef.model_construct(id=ref.id)


# ToDo: Use new syntax when requires-python >= 3.12
GO_co = TypeVar('GO_co', bound=GenericObject, default=GenericObject, covariant=True)


class User(TypedObject[GO_co], UserRef, MentionMixin, polymorphic_base=True):
    """Represents a User in Notion."""

    name: str | None = None
    avatar_url: str | None = None

    def build_mention(self, style: Annotations | None = None) -> MentionObject:
        return MentionUser.build_mention_from(self, style=style)


class PersonTypeData(GenericObject):
    """Type data for a `Person`."""

    email: str | None = None


class Person(User[PersonTypeData], type='person'):
    """Represents a Person in Notion."""

    person: PersonTypeData = Field(default_factory=PersonTypeData)


class WorkSpaceLimits(GenericObject):
    """Limits for a Notion workspace."""

    max_file_upload_size_in_bytes: int | None = None


class UnknownUserTypeData(GenericObject):
    """Type data for an `UnknownUser`."""


class UnknownUser(User[UnknownUserTypeData], type='unknown'):
    """Represents an unknown user in Notion.

    This is a unofficial placeholder for a user that is not recognized by the API.
    """

    name: Literal['Unknown User'] = 'Unknown User'
    unknown: UnknownUserTypeData = Field(default_factory=UnknownUserTypeData)


class Annotations(GenericObject):
    """Style information for RichTextObject's."""

    bold: bool = False
    italic: bool = False
    strikethrough: bool = False
    underline: bool = False
    code: bool = False
    # `Unset` for instance for code blocks to have flexible coloring. Will be replaced by Color.DEFAULT.
    color: Color | BGColor | UnsetType = Unset


MAX_TEXT_OBJECT_SIZE = 2_000
"""The max text size according to the Notion API is 2000 characters."""


class RichTextBaseObject(TypedObject[GenericObject], polymorphic_base=True):
    """Base class for Notion rich text elements."""

    plain_text: str
    href: str | None = None
    annotations: Annotations | UnsetType = Unset


def rich_text_to_str(rich_texts: list[RichTextBaseObject]) -> str:
    """Convert a list of rich texts to plain text."""
    return ''.join(rich_text.plain_text for rich_text in rich_texts)


class TextObject(RichTextBaseObject, type='text'):
    """Notion text element."""

    class TypeData(GenericObject):
        content: str
        link: LinkObject | None = None

    text: TypeData

    @classmethod
    def build(cls, text: str, *, href: str | None = None, style: Annotations | None = None) -> TextObject:
        """Compose a TextObject from the given properties.

        Args:
            text: the plain text of this object
            href: optional link for this object
            style: optional annotations for styling this text
        """
        link = LinkObject(url=href) if href else None
        nested = cls.TypeData(content=text, link=link)

        return cls.model_construct(
            plain_text=text,
            text=nested,
            href=href,
            annotations=Unset if style is None else deepcopy(style),
        )


class EquationObject(RichTextBaseObject, type='equation'):
    """Notion equation element."""

    class TypeData(GenericObject):
        expression: str

    equation: TypeData

    @classmethod
    def build(cls, expression: str, *, href: str | None = None, style: Annotations | None = None) -> EquationObject:
        """Compose a TextObject from the given properties.

        Args:
            expression: expression
            href: optional link for this object
            style: optional annotations for styling this text
        """
        return cls.model_construct(
            plain_text=expression,
            equation=cls.TypeData(expression=expression),
            href=href,
            annotations=Unset if style is None else deepcopy(style),
        )


class MentionBase(TypedObject[GO_co], ABC, polymorphic_base=True):
    """Base class for typed `Mention` objects.

    Note that this class is different to `MentionMixin`, which is used to
    provide a `build_mention` method for objects that can be mentioned in Notion.
    Here, we have a *class method* to build a mention object *from* the target object.
    """

    @classmethod
    @abstractmethod
    def build_mention_from(cls, *args: Any, **kwargs: Any) -> MentionObject:
        """Build a mention object for this type of mention from the actual target object."""


class MentionObject(RichTextBaseObject, type='mention'):
    """Notion mention element."""

    mention: SerializeAsAny[MentionBase]


class MentionUser(MentionBase[User | UserRef], type='user'):
    """Nested user data for `Mention` properties."""

    user: SerializeAsAny[User | UserRef]

    @classmethod
    def build_mention_from(cls, user: User, *, style: Annotations | None = None) -> MentionObject:
        # When creating a user mention, we need to send only the id, not the full object.
        user_ref = UserRef.build(user)
        mention = cls.model_construct(user=user_ref)
        # note that `href` is always `None` for user mentions, also we prepend the '@' to mimic server side
        return MentionObject.model_construct(
            plain_text=f'@{user.name}',
            href=None,
            annotations=Unset if style is None else deepcopy(style),
            mention=mention,
        )


class MentionLinkTypeData(GenericObject):
    """Type data for a `MentionLink`."""

    href: str
    title: str


class MentionLink(MentionBase, type='link_mention'):
    """Nested url data for `Mention` properties."""

    link_mention: MentionLinkTypeData

    @classmethod
    def build_mention_from(cls, href: str, title: str, *, style: Annotations | None = None) -> MentionObject:
        """Build a mention object for this type of mention from the actual target object."""
        mention = cls.model_construct(link_mention=MentionLinkTypeData(href=href, title=title))
        return MentionObject.model_construct(
            plain_text=title,
            href=None,
            annotations=Unset if style is None else deepcopy(style),
            mention=mention,
        )


class BotTypeData(GenericObject):
    """Type data for a `Bot`."""

    owner: WorkspaceRef | MentionUser | None = None
    workspace_name: str | None = None
    workspace_limits: WorkSpaceLimits = Field(default_factory=WorkSpaceLimits)


class Bot(User[BotTypeData], type='bot'):
    """Represents a Bot in Notion."""

    # Even if stated otherwise in the docs, `bot` type data is optional and for instance
    # not present when a new page is created by a bot within a database with a `CreatedBy` Property.
    # For ease of use, we include a default instance of the bot type data.
    bot: BotTypeData = Field(default_factory=BotTypeData)


class MentionPage(MentionBase[ObjectRef], type='page'):
    """Nested page data for `Mention` properties."""

    page: SerializeAsAny[ObjectRef]

    @classmethod
    def build_mention_from(cls, page: Page, *, style: Annotations | None = None) -> MentionObject:
        page_ref = ObjectRef.build(page)
        mention = cls.model_construct(page=page_ref)
        # note that `href` is always `None` for page mentions
        return MentionObject.model_construct(
            plain_text=rich_text_to_str(page.title),
            href=None,
            annotations=Unset if style is None else deepcopy(style),
            mention=mention,
        )


class MentionDatabase(MentionBase[ObjectRef], type='database'):
    """Nested database information for `Mention` properties."""

    database: SerializeAsAny[ObjectRef]

    @classmethod
    def build_mention_from(cls, db: Database, *, style: Annotations | None = None) -> MentionObject:
        db_ref = ObjectRef.build(db)
        mention = cls.model_construct(database=db_ref)
        # note that `href` is always `None` for database mentions
        db_title = raise_unset(db.title)
        return MentionObject.model_construct(
            plain_text=rich_text_to_str(db_title),
            ref=None,
            annotations=Unset if style is None else deepcopy(style),
            mention=mention,
        )


class MentionDate(MentionBase[DateRange], type='date'):
    """Nested date data for `Mention` properties."""

    date: DateRange

    @classmethod
    def build_mention_from(cls, date_range: DateRange, *, style: Annotations | None = None) -> MentionObject:
        mention = cls.model_construct(date=date_range)
        # note that `href` is always `None` for date mentions
        return MentionObject.model_construct(
            plain_text=str(date_range),
            ref=None,
            annotations=Unset if style is None else deepcopy(style),
            mention=mention,
        )


class MentionLinkPreviewTypeData(GenericObject):
    """Type data for a `MentionLinkPreview`."""

    url: str


class MentionLinkPreview(MentionBase, type='link_preview'):
    """Nested url data for `Mention` properties.

    !!! warning

        Link previews cannot be created via the API.
    """

    link_preview: MentionLinkPreviewTypeData

    @classmethod
    def build_mention_from(cls, url: str, *, style: Annotations | None = None) -> MentionObject:
        """Build a mention object for this type of mention from the actual target object."""
        mention = cls.model_construct(link_preview=MentionLinkPreviewTypeData.model_construct(url=url))
        return MentionObject.model_construct(
            plain_text=url,
            href=None,
            annotations=Unset if style is None else deepcopy(style),
            mention=mention,
        )


class MentionTemplateData(TypedObject[GenericObject]):
    """Nested template data for `Mention` properties."""


class MentionTemplate(MentionBase[MentionTemplateData], type='template_mention'):
    """Nested template data for `Mention` properties."""

    template_mention: SerializeAsAny[MentionTemplateData]


class MentionTemplateDate(MentionTemplateData, type='template_mention_date'):
    """Nested date template data for `Mention` properties."""

    template_mention_date: str


class MentionTemplateUser(MentionTemplateData, type='template_mention_user'):
    """Nested user template data for `Mention` properties."""

    template_mention_user: str


class EmojiObject(TypedObject[GenericObject], type='emoji'):
    """A Notion emoji object.

    Within text an emoji is represented as unicode string.
    """

    emoji: str

    @classmethod
    def build(cls, emoji: str) -> EmojiObject:
        """Compose an EmojiObject from the given emoji string."""
        return EmojiObject.model_construct(emoji=emoji)


class CustomEmojiObjectTypeData(GenericObject):
    """Type data for a `CustomEmojiObject`."""

    id: UUID
    name: str
    url: str


class CustomEmojiObject(MentionBase[CustomEmojiObjectTypeData], MentionMixin, type='custom_emoji'):
    """A Notion custom emoji object.

    Within text a custom emoji is represented as a mention. For this
    reason there is no `MentionCustomEmoji` class, but the `CustomEmojiObject`
    itself can be used to build a mention object.
    """

    custom_emoji: CustomEmojiObjectTypeData

    @classmethod
    def build_mention_from(cls, custom_emoji: CustomEmojiObject, *, style: Annotations | None = None) -> MentionObject:
        mention = cls.model_construct(custom_emoji=custom_emoji.custom_emoji)
        # note that `href` is always `None` for custom emoji mentions
        return MentionObject.model_construct(
            plain_text=f':{custom_emoji.custom_emoji.name}:',
            href=None,
            annotations=Unset if style is None else deepcopy(style),
            mention=mention,
        )

    def build_mention(self, style: Annotations | None = None) -> MentionObject:
        """Build a mention object for this custom emoji."""
        return self.__class__.build_mention_from(self, style=style)


class FileObject(TypedObject[GO_co], polymorphic_base=True):
    """A Notion file object.

    Depending on the context, a FileObject may require a name (such as in the `Files` property) or may have a caption,
    for instance when used within a File block, which makes the object hierarchy complex. Thus, we simply allow `name`
    and `caption` to be optional. It is the responsibility of the caller to set `name` if required by the API.
    """

    name: str | None = None
    caption: list[SerializeAsAny[RichTextBaseObject]] | None = None


class HostedTypedata(GenericObject):
    """Type data for `HostedFile`."""

    url: str
    expiry_time: dt.datetime | None = None


class HostedFile(FileObject[HostedTypedata], type='file'):
    """A Notion file object."""

    file: HostedTypedata

    @classmethod
    def build(
        cls,
        url: str,
        *,
        name: str | None = None,
        caption: list[RichTextBaseObject] | None = None,
        expiry_time: dt.datetime | None = None,
    ) -> HostedFile:
        """Create a new `HostedFile` from the given URL."""
        return cls.model_construct(name=name, caption=caption, file=HostedTypedata(url=url, expiry_time=expiry_time))


class ExternalFileTypeData(GenericObject):
    """Type data for `ExternalFile`."""

    url: str


class ExternalFile(FileObject[ExternalFileTypeData], type='external'):
    """A Notion external file object."""

    external: ExternalFileTypeData

    @classmethod
    def build(
        cls, url: str, *, name: str | None = None, caption: list[RichTextBaseObject] | None = None
    ) -> ExternalFile:
        """Create a new `ExternalFile` from the given URL."""
        return cls.model_construct(name=name, caption=caption, external=ExternalFileTypeData(url=url))


class UploadedFileTypeData(GenericObject):
    """Type data for `UploadedFile`."""

    id: UUID


class UploadedFile(FileObject[UploadedFileTypeData], type='file_upload'):
    """A Notion uploaded file object. The result of completed FileUpload"""

    file_upload: UploadedFileTypeData

    @classmethod
    def build(cls, id: UUID) -> UploadedFile:  # noqa: A002
        """Create a new `UploadedFile` from the given ID."""
        return cls.model_construct(file_upload=UploadedFileTypeData(id=id))


class FileImportTypeData(GenericObject):
    """Type data for `FileImportSuccess` and `FileImportError`.

    For ease of use, the parameters of success and error are combined into a single class.
    """

    type: str | None = None
    code: str | None = None
    message: str | None = None
    parameter: str | None = None
    status_code: int | None = None


class FileImportSuccess(TypedObject[GenericObject], type='success'):
    """Result of a successful file import operation."""

    imported_time: dt.datetime | None = None
    success: FileImportTypeData


class FileImportError(TypedObject[GenericObject], type='error'):
    """Result of a failed file import operation."""

    imported_time: dt.datetime | None = None
    error: FileImportTypeData


class FileUpload(NotionObject, object='file_upload'):
    """A Notion file upload object.

    This object is used to handle the process of uploading a file to Notion.
    """

    class NumberOfParts(GenericObject):
        """Number of parts for the file upload."""

        total: int
        sent: int

    id: UUID
    created_time: dt.datetime
    last_edited_time: dt.datetime
    expiry_time: dt.datetime | None = None
    status: FileUploadStatus
    filename: str | None = None
    content_type: str | None = None
    content_length: int | None = None
    number_of_parts: NumberOfParts | None = None
    upload_url: str | None = None
    complete_url: str | None = None
    file_import_result: FileImportSuccess | FileImportError | None = None
    archived: bool  # undocumented but sent by the API
    created_by: User  # undocumented but sent by the API


class Comment(NotionEntity, object='comment'):
    """A Notion comment object."""

    class TypeData(GenericObject):
        type: str  # This is strangely always `integration`, even if set by a user
        resolved_name: str  # This always the workspace name

    discussion_id: UUID
    rich_text: list[RichTextBaseObject]
    display_name: TypeData
